#!/usr/bin/env python3
# coding: utf_8


"""
Module to serialize GFF files.
"""

from .parser import Parser
from .gfannotation import GFAnnotation, _attribute_definition
from sys import intern
import re


# This class has exactly how many attributes I need it to have
# pylint: disable=too-many-instance-attributes
from ..exceptions import InvalidParsingFormat


class GffLine(GFAnnotation):
    """Object which serializes a GFF line."""

    # The (?:;|$) means "match, but **do not capture**, either semicolon or end of the line.
    _attribute_pattern = re.compile(r"([^;]*)=([^$=]*)(?:;|$)")

    def __init__(self, line, my_line='', header=False):
        """
        Constructor method.
        :param line: the GFF line to be serialised
        :type line: (str|None)

        :param my_line: optional string to be passed along. For exotic uses of the constructor.
        :type my_line: str

        :param header: boolean flag that indicates whether the instance will be a header or not.
        :type header: bool
        """

        GFAnnotation.__init__(self, line, my_line, header=header)
        self.__set_name()  # Set the name

    def _parse_attributes(self):

        """
        Private method that parses the last field of the GFF line.
        :return:
        """

        attr = self._attr.rstrip().rstrip(";")
        infolist = self._attribute_pattern.findall(attr)
        if not infolist and attr:
            raise InvalidParsingFormat(
                "The attribute section of this line could not be parsed correctly. Is this a GFF3?\n{}".format(
                    self._line))

        attribute_order = [key for key, val in infolist if key not in ("Parent", "parent", "id", "ID", "Id")]
        attributes = dict((key, _attribute_definition(val)) for key, val in infolist)

        # Ensure that the "parent" attribute is a string
        if "Parent" in attributes:
            attributes["Parent"] = str(attributes["Parent"])
            self.parent = attributes["Parent"]
        elif "parent" in attributes:
            attributes["parent"] = str(attributes["parent"])
            self.parent = attributes["parent"]

        # Ensure that the "ID" attribute is a string
        if "ID" in attributes:
            attributes["ID"] = str(attributes["ID"])
            self.id = attributes["ID"]
        elif "id" in attributes:
            attributes["id"] = str(attributes["id"])
            self.id = attributes["id"]
        elif "Id" in attributes:
            attributes["Id"] = str(attributes["Id"])
            self.id = attributes["Id"]

        self.attributes.update(attributes)
        self.attribute_order = attribute_order

    def _format_attributes(self):
        """
        Implementation of the abstract method for formatting the
        ninth field of a GFF.
        :return:
        """

        attrs = self._format_attributes_dict(
            self.attributes,
            parent=self.parent,
            mid=self.id,
            name=self.name,
            attribute_order=self.attribute_order)

        return attrs

    @staticmethod
    def _format_attributes_dict(attributes, parent=None, mid=None, name=None, attribute_order=None):

        if not parent:
            try:
                parent = attributes["parent"]
            except KeyError:
                try:
                    parent = attributes["Parent"]
                except KeyError:
                    parent = None

        if not mid:
            try:
                mid = attributes["id"]
            except KeyError:
                try:
                    mid = attributes["ID"]
                except KeyError:
                    mid = None
        if not name:
            name = attributes.get("name", None)
            if not name:
                name = attributes.get("Name", None)
        attrs = []
        if mid is not None:
            attrs.append("ID={0}".format(mid))
        if parent is not None:
            if isinstance(parent, str):
                parent = [parent]
            if len(parent) > 1:
                attrs.append("Parent={0}".format(",".join(parent)))
            elif len(parent) == 1:
                attrs.append("Parent={0}".format(parent[0]))
        if name is not None:
            attrs.append("Name={0}".format(name))

        if not attribute_order:
            attribute_order = sorted(list(key for key in attributes if
                                               key not in ["ID", "Parent", "Name",
                                                           "parent", "id", "name"]))

        attrs += ["{0}={1}".format(att.lower(), attributes[att]) for att in attribute_order
                  if (att not in ("ID", "Parent", "Name", "name", "gene_id", "transcript_id") and
                      attributes[att] is not None)]

        attrs = ";".join(attrs)
        return attrs

    # id is and id remains
    # pylint: disable=invalid-name
    @property
    def id(self):
        """
        Returns the ID of the feature.
        :rtype str
        """
        if "ID" in self.attributes:
            return str(self.attributes["ID"])
        else:
            return None

    @id.setter
    def id(self, newid):
        """
        :param newid: new id of the feature
        :type newid: str
        Setter for instance id.
        """

        self.attributes["ID"] = newid
        self._is_gene = None
    # pylint: enable=invalid-name

    @property
    def parent(self):
        """This property looks up the "Parent" field in the "attributes" dictionary.
        Contrary to other attributes, this property returns a *list*, not a string.
        This is due to the fact that GFF files support
        multiple inheritance by separating the parent entries with a comma.

        :rtype list

        """
        return self.__parent

    def __set_parent(self):

        if "Parent" not in self.attributes:
            self.parent = None
        self.parent = self.attributes["Parent"]

    @parent.setter
    def parent(self, parent):
        """
        Setter for the parent attribute. It will be converted into a list if it is not already.

        :param parent: the new parent attribute
        :type parent: str
        :type parent: list
        """

        if isinstance(parent, str):
            parent = parent.split(",")
        self.attributes["Parent"] = parent
        self.__parent = parent

    @property
    def name(self):
        """
        Returns the name of the feature. It defaults to the ID if missing.

        :rtype str
        """

        return self.__name

    def __set_name(self):

        if "Name" not in self.attributes:
            self.name = None
        self.__name = self.attributes["Name"]

    @name.setter
    def name(self, name):
        """
        Setter for the name attribute.
        :param name
        :type name: str
        """

        if not isinstance(name, (type(None), str)):
            raise TypeError("Invalid value for name: {0}".format(name))
        self.attributes["Name"] = name
        self.__name = name

    @property
    def is_transcript(self):
        """
        Property. True if the feature is an RNA, false otherwise.
        :rtype bool
        """

        if self.feature is None:
            return False
        if self.is_gene is True and self.feature != "mRNA_TE_gene":
            return False
        if self.feature.endswith("transcript") or "RNA" in self.feature.upper():
            return True
        elif self.feature.endswith("_gene_segment"):  # Necessary for V_gene_segment, C_gene_segment, etc.
            return True
        return False

    @property
    def is_parent(self):
        """
        Property. True if the feature has no parent defined.
        :rtype bool
        """

        if len(self.parent) == 0:
            return True
        return False

    @property
    def gene(self):
        """
        Property. If the feature is a transcript,
        returns the parent; else, it returns None
        (as only transcript lines have the gene among the attributes)
        :rtype str
        """

        if self.is_transcript is True and self.parent:
            return self.parent[0]
        elif self.is_gene:
            return self.id
        else:
            return None

    @property
    def transcript(self):
        """
        Property. If the feature is a transcript, it returns the id;
        if it is an exon, it returns the parent; else, it returns None

        :rtype str
        :rtype list
        :rtype None
        """
        if self.is_exon is True:
            return self.parent
        elif self.is_transcript is True:
            return self.id
        else:
            return None

    @transcript.setter
    def transcript(self, string):
        """
        Setter for the transcript property. It performs a light sanity check on the new
        assigned value.
        :param string:
        :return:
        """
        if self.is_exon is True:
            self.parent = string
        elif self.is_transcript is True:
            self.id = string
        else:
            raise TypeError("Cannot set the transcript name of a non-transcript feature!")

    @property
    def is_derived(self):
        """
        Property. It checks whether there is a "Derives_from" attribute among the line attributes.
        :rtype bool
        """
        return any(("derives_from" in _.lower()) for _ in self.attributes.keys())

    @property
    def derived_from(self):
        """
        Property, Set to True if the key "Derives_from" is present among the
        instance attributes.
        """

        if self.is_derived is False:
            return None
        else:
            if "Derives_from" not in self.attributes:
                key = [_ for _ in self.attributes.keys() if _.lower() == "derives_from"]
                assert len(key) == 1, (str(self), "Multiple 'derives_from' found:\n"+ str(key))
                self.attributes["Derives_from"] = self.attributes[key[0]]

            return self.attributes["Derives_from"].split(",")

    @property
    def _negative_order(self):
        negative_order = [intern(_) for _ in ["three_prime_utr",
                                              "exon",
                                              "stop_codon",
                                              "CDS",
                                              "start_codon",
                                              "five_prime_utr"]]
        return negative_order

    @property
    def _positive_order(self):
        positive_order = [intern(_) for _ in
                          ["five_prime_utr",
                           "exon",
                           "start_codon",
                           "CDS",
                           "stop_codon",
                           "three_prime_utr"]]
        return positive_order


class GFF3(Parser):
    """
    Class that is used to parse a GFF file.
    """

    __annot_type__ = "gff3"

    def __init__(self, handle):
        """
        Constructor method.
        :param handle: the input file. It can be a file handle or a file name.
        :type handle: io.TextIOWrapper | str
        """
        super().__init__(handle)
        self.header = False
        self.__line_counter = 0

    def __next__(self):

        if self.closed:
            raise StopIteration
        try:
            line = next(self._handle)
            self.__line_counter += 1
        except (ValueError, KeyError, TypeError, UnicodeError, AttributeError, AssertionError) as exc:
            line = None
            error = "Invalid line for file {name}, position {counter}:\n{line}Error: {exc}".format(
                name=self.name, counter=self.__line_counter, line=line, exc=exc)
            raise InvalidParsingFormat(error)

        if line[0] == "#":
            return GffLine(line, header=True)

        try:
            gff_line = GffLine(line)
        except Exception as exc:
            error = "Invalid line for file {name}, position {counter}:\n{line}Error: {exc}".format(
                name=self.name, counter=self.__line_counter, line=line, exc=exc)
            raise InvalidParsingFormat(error)
        return gff_line

    @property
    def file_format(self):
        return self.__annot_type__



import unittest
# import Mikado
from .. import utilities, configuration, serializers
import tempfile
import sqlalchemy.orm
import pandas as pd
import numpy as np
from ..utilities.dbutils import DBBASE
from ..serializers.external import External, ExternalSource, ExternalSerializer
from ..serializers.orf import Query
import sqlalchemy.exc


__author__ = 'Luca Venturini'


class TestExternal(unittest.TestCase):

    logger = utilities.log_utils.create_null_logger("test_external_serialise")

    def setUp(self):
        self.dbfile_handle = tempfile.NamedTemporaryFile(suffix=".db")
        self.dbfile = self.dbfile_handle.name
        self.configuration = configuration.configurator.load_and_validate_config(None)
        self.configuration.db_settings.dbtype = "sqlite"
        self.configuration.db_settings.db = self.dbfile
        self.__create_session()

    def tearDown(self) -> None:
        self.dbfile_handle.close()

    def __create_session(self):
        self.engine = utilities.dbutils.connect(
            self.configuration, self.logger)
        self.sessionmaker = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = self.sessionmaker()
        DBBASE.metadata.create_all(self.engine)

    def test_serialize_source(self):

        source = ExternalSource("foo", np.dtype("float"), False)
        self.session.add(source)
        self.session.commit()

        self.assertEqual(self.session.query(serializers.external.ExternalSource).count(), 1)

    def test_add_score(self):

        source = ExternalSource("cdna_length", np.dtype("int"), False)
        query = Query("foo.1", 200)
        self.session.add(source)
        self.session.add(query)
        self.session.commit()

        score = External(query.query_id, source.source_id, 200)
        self.session.add(score)

        # Before commiting, everything is detached
        self.assertIs(score.source, None)
        self.session.commit()

        self.assertEqual(score.source, source.source)
        self.assertEqual(score.query, query.query_name)
        self.assertEqual(self.session.query(serializers.external.External).count(), 1)

    def test_wrong_score(self):
        source = ExternalSource("cdna_length", np.dtype("int"), False)
        query = Query("foo.1", 200)
        self.session.add(source)
        self.session.add(query)
        self.session.commit()

        # This fails because the query_id
        score = External(query.query_id + 1, source.source_id, 200)
        self.session.add(score)
        with self.assertRaises(sqlalchemy.exc.IntegrityError):
            self.session.commit()
        self.session.rollback()

        score = External(query.query_id, source.source_id + 1, 200)
        self.session.add(score)
        with self.assertRaises(sqlalchemy.exc.IntegrityError):
            self.session.commit()
        self.session.rollback()

        with self.assertRaises(sqlalchemy.exc.ArgumentError):
            _ = External(query_id=query.query_id, source_id=source.source_id, score="200")

    def test_load_table(self):

        queries = ["T{}".format(_) for _ in range(1, 20)]
        [self.session.add(Query(_, 100)) for _ in queries]
        self.session.commit()

        df = pd.DataFrame({'tid': queries,
                           'foo': [100 for _ in queries],
                           'bar': [10 for _ in queries]})
        df.set_index("tid", inplace=True)
        with tempfile.NamedTemporaryFile() as temp_df:
            df.to_csv(temp_df.name, sep="\t", index_label="tid", mode="wt", header=True)
            temp_df.flush()
            conf = self.configuration.copy()
            conf.serialise.files.transcripts = None
            serializer = ExternalSerializer(temp_df.name, conf)
            serializer.serialize()



# cython: infer_types=True
# cython: language_level=3
# distutils: language=c++
import re
import numpy as np
cimport numpy as np
from libcpp.string cimport string
cimport cython

btop_pattern = re.compile(r"(\d+|\D{2,2})")

@cython.wraparound(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef parse_btop(str btop, Py_ssize_t qpos, Py_ssize_t spos,
                 np.ndarray[dtype=long, ndim=2, cast=True] query_array,
                 np.ndarray[dtype=long, ndim=2, cast=True] target_array,
                 dict matrix, long qmult=3, long tmult=1):

    """Parse the BTOP lines of tabular BLASTX/DIAMOND output.
    In BLASTX, the alignment *never* skips bases. Ie the relationship is *always* 3 bases to 1 aminoacid,
    even when there are gaps. It is therefore possible to perfectly infer the position of identical vs positive
    matches.

    The function expects the starting position of the match, the

    """

    # 0: match
    # 1: identical
    # 2: positive

    cdef str pos
    cdef long ipos
    cdef long qstart = qpos
    cdef long sstart = spos

    cdef long[:,:] query_view = query_array
    cdef long[:,:] target_view = target_array
    cdef long aln_span = 0
    cdef long step = min(qmult, tmult)
    cdef string match
    cdef Py_ssize_t idx

    for pos in btop_pattern.findall(btop):
        try:
            ipos = int(pos)
        except ValueError:
            if len(pos) != 2:
                raise ValueError(pos)
            ipos = - 1
        if ipos >= 0:
            aln_span += step * ipos
            for idx in range(step * ipos):
                match.push_back(b"|")
            query_view[:, qpos:qpos + ipos * qmult] = 1
            target_view[:, spos:spos + ipos * tmult] = 1
            qpos += ipos * qmult
            spos += ipos * tmult
        else:
            aln_span += step
            if pos[0] == "-":  # Gap in query
                target_view[0, spos:spos + tmult] = 1
                spos += tmult
                match.push_back(b"-")
            elif pos[1] == "-":  # Gap in target
                query_view[0, qpos:qpos + qmult] = 1
                qpos += qmult
                if pos[0] == "*":
                    match.push_back(b"*")
                else:
                    match.push_back(b"_")
            else:
                query_view[0, qpos:qpos + qmult] = 1
                target_view[0, spos:spos + tmult] = 1
                if matrix.get(pos, -1) > 0:
                    query_view[2, qpos:qpos + qmult] = 1
                    target_view[2, spos:spos + tmult] = 1
                    match.push_back(b"+")
                else:
                    if pos[0] == "*" or pos[1] == "*":
                        match.push_back(b"*")
                    else:
                        match.push_back(b"/")

                qpos += qmult
                spos += tmult

    fmatch = <bytes>match
    fmatch = fmatch.decode()

    return query_array, target_array, aln_span, fmatch



import functools
import multiprocessing
import multiprocessing.queues
import os
import pysam
import msgpack
from ..transcripts.transcriptchecker import TranscriptChecker
from .. import exceptions
from ..transcripts import Transcript
from ..utilities.log_utils import create_null_logger, create_queue_logger
import logging
import queue
import time
import sys
import operator
import random
import zlib


__author__ = 'Luca Venturini'


def create_transcript(lines,
                      fasta_seq,
                      start,
                      end,
                      lenient=False,
                      is_reference=False,
                      strand_specific=False,
                      canonical_splices=(("GT", "AG"),
                                         ("GC", "AG"),
                                         ("AT", "AC")),
                      strip_faulty_cds=False,
                      logger=None):
    """Function to create the checker.

    :param lines: all the exon lines for an object
    :type lines: dict

    :param fasta_seq: genomic sequence of the transcript

    :param start: start position for the transcript
    :type start: int
    :param end: end position for the transcript
    :type end: int

    :type lenient: bool
    :type strand_specific: bool

    :param canonical_splices: the splices considered as canonical for the species.
    :type canonical_splices: list[tuple]

    :param strip_faulty_cds: boolean. If set to true, transcripts with invalid CDSs or on the wrong strand will have
    their CDS removed. Otherwise, in such cases, they will be discarded by raising an InvalidTranscript exception.
    :type strip_faulty_cds: bool

    :param logger: optional logger to use during processing.

    :param is_reference: boolean. If set, the transcript's strand will not be checked.


    :rtype: (None|TranscriptChecker)
    """

    if logger is None:
        logger = create_null_logger()

    if "tid" not in lines:
        logger.error("Lines datastore lacks the transcript ID. Exiting.")
        return None

    try:
        logger.debug("Starting with %s", lines["tid"])
        transcript_line = Transcript()
        transcript_line.chrom = lines["chrom"]
        if "source" in lines:
            transcript_line.source = lines["source"]
        transcript_line.strand = lines["strand"]
        transcript_line.attributes.update(lines["attributes"])
        transcript_line.feature = "transcript"
        transcript_line.start, transcript_line.end = sorted([start, end])
        transcript_line.logger = logger
        assert lines["tid"] is not None, lines
        transcript_line.id = lines["tid"]
        transcript_line.parent = lines["parent"]

        for feature in lines["features"]:
            coords, phases = [], []
            for feat in lines["features"][feature]:
                try:
                    assert isinstance(feat, (list, tuple)) and 2 <= len(feat) <= 3, feat
                except AssertionError:
                    raise exceptions.InvalidTranscript("Invalid feature")
                coords.append((feat[0], feat[1]))
                if len(feat) == 3 and feat[2] in (0, 1, 2, None):
                    phases.append(feat[2])
                else:
                    phases.append(None)
            try:
                assert len(phases) == len(coords)
            except AssertionError:
                raise exceptions.InvalidTranscript("Invalid phases/coords")
            transcript_line.add_exons(coords, features=feature, phases=phases)

        transcript_object = TranscriptChecker(transcript_line,
                                              fasta_seq,
                                              lenient=lenient,
                                              strand_specific=strand_specific,
                                              canonical_splices=canonical_splices,
                                              strip_faulty_cds=strip_faulty_cds,
                                              is_reference=is_reference,
                                              logger=logger)
        logger.debug("Finished adding exon lines to %s", lines["tid"])
        transcript_object.finalize()
        transcript_object.check_strand()
        transcript_object.check_orf()
    except exceptions.IncorrectStrandError:
        logger.info("Discarded %s because of incorrect fusions of splice junctions",
                    lines["tid"])
        # logger.exception(exc)
        transcript_object = None
    except exceptions.InvalidTranscript as exc:
        logger.info("Discarded generically invalid transcript %s, exception: %s",
                    lines["tid"], exc)
        transcript_object = None
    except AssertionError as exc:
        logger.info("Validation failed on %s, assertion failure: %s",
                    lines["tid"], exc)
        transcript_object = None
    except KeyboardInterrupt:
        raise KeyboardInterrupt
    except Exception as exc:
        logger.exception(exc)
        transcript_object = None

    return transcript_object


class CheckingProcess(multiprocessing.Process):

    def __init__(self,
                 batch_file,
                 logging_queue,
                 fasta,
                 identifier,
                 shelve_stacks,
                 fasta_out,
                 gtf_out,
                 tmpdir,
                 lenient=False,
                 seed=None,
                 strip_faulty_cds=False,
                 canonical_splices=(("GT", "AG"),
                                    ("GC", "AG"),
                                    ("AT", "AC")),
                 log_level="WARNING"
                 ):

        super().__init__()
        self.__identifier = ""
        self.__set_identifier(identifier)
        if seed is not None:
            # numpy.random.seed(seed % (2 ** 32 - 1))
            random.seed(seed % (2 ** 32 - 1))
        else:
            # numpy.random.seed(None)
            random.seed(None)
        # self.strand_specific = strand_specific
        self.__canonical = []
        self.__set_canonical(canonical_splices)
        self.__log_level = "DEBUG"
        self.log_level = log_level
        self.logger = None  # This gets populated by the create_queue_logger function below
        self.__logging_queue = None
        self.__set_logging_queue(logging_queue)
        self.name = "Checker-{0}".format(self.identifier)
        try:
            create_queue_logger(self)
        except AttributeError as exc:
            raise AttributeError(exc)
        if batch_file is None:
            raise ValueError("Invalid Batch file!")
        self.__lenient = False
        self.lenient = lenient
        self.__fasta = fasta
        self.__submission_queue = None
        self.fasta = pysam.FastaFile(self.__fasta)
        self.fasta_out = os.path.join(tmpdir, "{0}-{1}".format(
            fasta_out, self.identifier
        ))
        self.gtf_out = os.path.join(tmpdir, "{0}-{1}".format(
            gtf_out, self.identifier
        ))
        self.strip_faulty_cds= strip_faulty_cds
        self.shelve_stacks = shelve_stacks
        self.batch_file = batch_file

    def _get_stacks(self):
        shelve_stacks = dict()
        for shelf in self.shelve_stacks:
            shelve_stacks[shelf] = open(shelf, "rb")

        return shelve_stacks

    def _get_keys(self):
        keys = msgpack.loads(open(self.batch_file, "rb").read(), raw=False, strict_map_key=False)
        keys = sorted(keys, key=operator.itemgetter(0))
        return keys

    def run(self):
        checker = functools.partial(create_transcript,
                                    # lenient=self.lenient,
                                    strip_faulty_cds=self.strip_faulty_cds,
                                    canonical_splices=self.canonical,
                                    logger=self.logger)

        fasta_out = open(self.fasta_out, "w")
        gtf_out = open(self.gtf_out, "w")
        self.logger.debug("Starting %s", self.name)
        self.logger.debug("Created output FASTA {self.fasta_out} and GTF {self.gtf_out}".format(**locals()))
        time.sleep(0.1)
        self.logger.debug(self.canonical)

        __printed = 0
        shelve_stacks = self._get_stacks()
        file_keys = self._get_keys()

        try:
            for key in file_keys:
                counter, keys = key
                # lines, start, end, counter = self.submission_queue.get()
                tid, chrom, (pos) = keys
                try:
                    tid, shelf_name, write_start, write_length = tid
                except ValueError as exc:
                    raise ValueError(f"{exc}\t{tid}")
                start, end = pos
                try:
                    shelf = shelve_stacks[shelf_name]
                except KeyError:
                    exception = f"{shelf_name} not found in shelves, available: {shelve_stacks.keys()}"
                    self.logger.error(exception)
                    raise KeyError(exception)

                shelf.seek(write_start)
                lines = msgpack.loads(zlib.decompress(shelf.read(write_length)))

                self.logger.debug("Checking %s", lines["tid"])
                if "is_reference" not in lines:
                    raise KeyError(lines)

                transcript = checker(lines,
                                     str(self.fasta.fetch(lines["chrom"], start-1, end)),
                                     start,
                                     end,
                                     lenient=self.lenient,
                                     is_reference=lines["is_reference"],
                                     strand_specific=lines["strand_specific"])

                if transcript is None:
                    self.logger.debug("%s failed the check", lines["tid"])
                    continue
                else:
                    self.logger.debug("Printing %s", lines["tid"])
                    __printed += 1
                    print("\n".join(["{0}/{1}".format(counter, line) for line in
                                     transcript.format("gtf").split("\n")]), file=gtf_out)
                    print("\n".join(["{0}/{1}".format(counter, line) for line in
                                     transcript.fasta.split("\n")]), file=fasta_out)
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except Exception as exc:
            self.logger.error(exc)
            # self.logging_queue.close()
            raise

        time.sleep(0.1)
        fasta_out.flush()
        fasta_out.close()
        gtf_out.flush()
        gtf_out.close()
        if __printed > 0:
            self.logger.debug("Size of FASTA out and GTF out: %s, %s",
                              os.stat(fasta_out.name).st_size, os.stat(gtf_out.name).st_size)
            assert os.stat(gtf_out.name).st_size > 0
            assert os.stat(fasta_out.name).st_size > 0
        time.sleep(0.1)
        return

    def __getstate__(self):
        state = self.__dict__.copy()
        for key in ("fasta", "logger", "_log_handler"):
            if key in state:
                del state[key]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        create_queue_logger(self)
        self.fasta = pysam.FastaFile(self.__fasta)

    @property
    def identifier(self):
        return self.__identifier

    def __set_identifier(self, identifier):

        if identifier is None:
            raise ValueError("The identifier must be defined!")
        self.__identifier = str(identifier)

    @property
    def log_level(self):
        return self.__log_level

    @log_level.setter
    def log_level(self, log_level):
        _ = logging._checkLevel(log_level)
        self.__log_level = log_level

    @property
    def lenient(self):
        return self.__lenient

    @lenient.setter
    def lenient(self, lenient):
        if lenient not in (False, True):
            raise ValueError("Invalid lenient value: {}".format(lenient))
        self.__lenient = lenient

    # @property
    # def submission_queue(self):
    #     return self.__submission_queue
    #
    # def __set_submission_queue(self, submission):
    #     if isinstance(submission, multiprocessing.queues.SimpleQueue):
    #         if sys.version_info.minor < 6:
    #             raise TypeError("Invalid queue object for Python 3.5 and earlier!")
    #         submission.put_nowait = submission.put
    #     elif not isinstance(submission, (multiprocessing.queues.Queue,
    #                                    queue.Queue)):
    #         raise TypeError("Invalid queue object: {}".format(type(submission)))
    #     self.__submission_queue = submission
    #
    @property
    def logging_queue(self):
        return self.__logging_queue

    def __set_logging_queue(self, logging_queue):
        if isinstance(logging_queue, multiprocessing.queues.SimpleQueue):
            if sys.version_info.minor < 6:
                raise TypeError("Invalid queue object for Python 3.5 and earlier!")
            logging_queue.put_nowait = logging_queue.put
        elif not isinstance(logging_queue, (multiprocessing.queues.Queue, queue.Queue)):
            raise TypeError("Invalid queue object: {}".format(type(logging_queue)))
        self.__logging_queue = logging_queue

    @property
    def canonical(self):
        return self.__canonical

    def __set_canonical(self, canonical):
        if not isinstance(canonical, (tuple, list)):
            raise TypeError("Canonical splices should be lists or tuples")

        if len(canonical) == 0:
            raise ValueError("The list of canonical splices cannot be empty!")

        for el in canonical:
            if (len(el) != 2 or (not (isinstance(el[0], str) and len(el[0]) == 2) or
                                not (isinstance(el[1], str) and len(el[1]) == 2  ))):
                raise ValueError("Invalid splicing pattern!")

        self.__canonical = canonical



"""
This module defines the functions needed to check the sanity of the configuration file,
plus the JSON schemas for the configuration and scoring files.
"""
from typing import Union

import os
from marshmallow import fields
from .configuration import MikadoConfiguration
from .daijin_configuration import DaijinConfiguration
from . import configurator
import itertools
import textwrap
import rapidjson as json
import toml
import dataclasses
import yaml


__author__ = 'Luca Venturini'


def print_config(config: Union[MikadoConfiguration, DaijinConfiguration], out,
                 output_format="yaml", no_files=False, full=True):
    """
    Function to print out the configuration in TOML format, adding the descriptions as comments preceded by #.
    :param config: configuration
    :type config: (MikadoConfiguration|DaijinConfiguration)

    :param out: output handle
    :type out: [io.TextIOWrapper|io.TextIO]

    :param output_format: one of yaml, json or toml (case-insensitive)
    :type output_format: str

    :param no_files: boolean. If on, sections pertaining to files will be deleted from the output.
    :type no_files: bool
    """

    if not isinstance(output_format, str) or output_format.lower() not in ("yaml", "json", "toml"):
        raise ValueError("Unknown format: {}. I can only accept yaml, json or toml as options.")

    output_format = output_format.lower()
    config_dict = dataclasses.asdict(config)
    for key in ["scoring", "cds_requirements", "requirements", "not_fragmentary", "as_requirements"]:
        # Necessary otherwise we will be deleting fields from the *original* object!
        config_dict.pop(key, None)
    if no_files is True:
        for stage in ["pick", "prepare", "serialise"]:
            if "files" in config_dict[stage]:
                del config_dict[stage]["files"]
        del config_dict["reference"]
        del config_dict["db_settings"]

    if not full:
        config_dict = filter_config(config_dict, config)

    if output_format == "toml":
        print_toml_config(config_dict, config, out)
    elif output_format == "yaml":
        print_yaml_config(config_dict, config, out)
    elif output_format == "json":
        # Necessary otherwise we will be deleting fields from the *original* object!
        print(json.dumps(config_dict, indent=4, sort_keys=True), file=out)


def select_attribute_for_output(final_config_level: dict, attr_parent, attr_name, attr_value):
    is_nested = isinstance(attr_parent.Schema._declared_fields[attr_name], fields.Nested)
    is_required = attr_parent.Schema._declared_fields[attr_name].required

    # This is not going to happen if we are looking at a nested property
    if attr_parent.Schema._declared_fields[attr_name].dump_default == attr_value:
        if is_required:
            final_config_level[attr_name] = attr_value
        else:
            return
    elif not is_nested:
        callable_and_equal_to_default = (callable(attr_parent.Schema._declared_fields[attr_name].dump_default) and
                                         attr_parent.Schema._declared_fields[attr_name].dump_default() == attr_value)
        if callable_and_equal_to_default and not is_required:
            return
        else:
            final_config_level[attr_name] = attr_value
    else:
        final_config_level[attr_name] = dict()
        for key, value in dataclasses.asdict(attr_value).items():
            select_attribute_for_output(
                final_config_level[attr_name], attr_value, key, getattr(attr_value, key))


def clear_dict(data: dict) -> dict:
    keys = list(data.keys())
    for key in keys:
        if isinstance(data[key], dict) and len(data[key]) > 0:
            data[key] = clear_dict(data[key])
        if isinstance(data[key], dict) and len(data[key]) == 0:
            del data[key]
        else:
            pass
    return data


def filter_config(config_dict: dict, config: Union[DaijinConfiguration, MikadoConfiguration]):
    for key in ["scoring", "cds_requirements", "requirements", "not_fragmentary", "as_requirements"]:
        config_dict.pop(key, None)
    filtered_config = dict()

    # Recursive method here
    for key, value in config_dict.items():
        orig_type = type(value)
        select_attribute_for_output(filtered_config, config, key, getattr(config, key))
    clear_dict(filtered_config)
    return filtered_config


def print_toml_config(config_dict: dict, config: Union[DaijinConfiguration, MikadoConfiguration], out):

    """Function to print out the configuration in TOML format, adding the descriptions as comments preceded by #.
    :param config_dict: the configuration dictionary, optionally filtered
    :type config_dict: dict

    :param config: configuration object
    :type config: (MikadoConfiguration|DaijinConfiguration)

    :param out: output handle
    :type out: io.TextIOWrapper

    :param no_files: boolean. If on, sections pertaining to files will be deleted from the output.
    :type no_files: bool
    """

    output = toml.dumps(config_dict)

    lines = []
    level = config

    for line in output.split("\n"):
        if line.startswith("["):
            keys = line.rstrip().replace("[", "").replace("]", "").split(".")
            level = config
            for key in keys[:-1]:
                level = getattr(level, key)
            metadata = level.Schema._declared_fields[keys[-1]].metadata
            comment = []
            description = metadata.get('description', None)
            if description:
                comment += ["# " + _ for _ in textwrap.wrap(description)]
            dataclassObject = getattr(level, keys[-1])
            if hasattr(dataclassObject, "Schema"):
                level = getattr(level, keys[-1])
            lines.append(line)
            lines.extend(comment)
        else:
            comment = []
            if "=" in line:
                key = line.split("=")[0].strip()
                meta = level.Schema._declared_fields.get(key)

                description = None
                if meta:
                    description = meta.metadata.get("description")
                if description:
                    _comment = textwrap.wrap(description)
                    if _comment:
                        _comment[0] = key + ": " + _comment[0]
                        comment += ["# " + _ for _ in _comment]
                lines.extend(comment)
            # if level.Schema._declared_fields[key].required or full:
            lines.append(line.rstrip())

    if config.__doc__:
        print(*["# " + _ for _ in textwrap.wrap(config.__doc__.strip())], sep="\n", file=out)
        print("#", file=out)

    print(*lines, sep="\n", file=out)


def print_yaml_config(config_dict: dict, config: Union[DaijinConfiguration, MikadoConfiguration], out):
    """
    Function to print out the configuration in YAML format, adding the descriptions as comments preceded by #.
    :param config_dict: the configuration dictionary, optionally filtered
    :type config_dict: dict

    :param config: configuration object
    :type config: (MikadoConfiguration|DaijinConfiguration)

    :param out: output handle
    :type out: io.TextIOWrapper

    :param no_files: boolean. If on, sections pertaining to files will be deleted from the output.
    :type no_files: bool
    """

    output = yaml.dump(config_dict, default_flow_style=False)
    lines = []
    nesting = []
    comment = []

    for line in output.split("\n"):
        if line and line.startswith(line.lstrip()[0]):  # We are back at the head level
            nesting = []
        if line.endswith(":"):  # New level
            spaces = sum(1 for _ in itertools.takewhile(str.isspace, line))
            key = line.strip().rstrip(":")
            while nesting and nesting[-1][1] >= spaces:
                nesting = nesting[:-1]
            level = config
            for oldkey, _ in nesting:
                _ = getattr(level, oldkey)
                level = _
            try:
                field_dc = level.Schema._declared_fields[key]
            except KeyError:
                raise KeyError(key, level)
            except AttributeError:
                raise AttributeError(key, level)
            comment = []
            description = field_dc.metadata.get("description", None)
            if description:
                comment += ["# " + _ for _ in textwrap.wrap(description)]
            lines.append(line)
            nesting.append((key, spaces))
        else:
            spaces = sum(1 for _ in itertools.takewhile(str.isspace, line))
            comment = [" " * spaces + _ for _ in comment]
            lines.extend(comment)
            comment = []
            key = line.split(":")[0].strip()
            level = config
            if ":" in line:
                for stop, (nest_level, _) in enumerate(nesting):
                    _ = getattr(level, nest_level)
                    if not hasattr(_, "Schema"):
                        break
                    elif key in _.Schema._declared_fields:
                        level = _
                        break
                    level = _
                try:
                    meta = level.Schema._declared_fields[key].metadata
                    description = meta.get("description", None)
                except AttributeError:
                    raise AttributeError(key, level)
                except KeyError:
                    description = None

                if description:
                    _comment = textwrap.wrap(description)
                    if _comment:
                        _comment[0] = key + ": " + _comment[0]
                        comment += [" " * spaces + "# " + _ for _ in _comment]

            lines.extend(comment)
            lines.append(line.rstrip())
            comment = []

    if config.__doc__:
        print(*["# " + _ for _ in textwrap.wrap(config.__doc__.strip())], sep="\n", file=out)
        print("#", file=out)

    print(*lines, sep="\n", file=out)


def parse_list_file(cfg: Union[MikadoConfiguration,DaijinConfiguration],
                    list_file: str) -> Union[MikadoConfiguration,DaijinConfiguration]:
    configuration = {
        "pick": {
            "chimera_split": {
                "skip": []
            }
        },
        "prepare": {
            "files": {
                "gff": [],
                "labels": [],
                "strand_specific_assemblies": [],
                "source_score": {},
                "reference": [],
                "exclude_redundant": [],
                "strip_cds": [],
                "skip": []
            }
        }
    }

    if isinstance(list_file, str):
        list_file = open(list_file)

    for counter, line in enumerate(list_file):
        line = line.strip()
        if not line or line.startswith("#"):
            continue  # Skip empty lines or headers
        fields = line.rstrip().split("\t")
        try:
            gff_name, label, stranded = fields[:3]
        except (IndexError, ValueError):
            raise IndexError(
                "The input list file must contain, for each line, at a minimum file name, unique label and "
                "strandedness. Line no. {} does not contain the expected fields.".format(counter))
        if not os.path.exists(gff_name):
            raise ValueError("Non-existent file name at line {}: {}".format(counter, gff_name))
        if label in configuration["prepare"]["files"]["labels"]:
            raise ValueError("Non-unique label specified at line {}: {}".format(counter, label))
        if stranded.lower() not in ("true", "false"):
            raise ValueError("Malformed line (no. {}) for the list: {}".format(counter, line))
        if gff_name in configuration["prepare"]["files"]["gff"]:
            raise ValueError("Repeated input file at line no. {}: {}".format(counter, line))
        elif label != '' and label in configuration["prepare"]["files"]["labels"]:
            raise ValueError("Repeated label at line {}: {}".format(counter, line))
        configuration["prepare"]["files"]["gff"].append(gff_name)
        configuration["prepare"]["files"]["labels"].append(label)
        if stranded.capitalize() == "True":
            configuration["prepare"]["files"]["strand_specific_assemblies"].append(gff_name)
        elif stranded.capitalize() == "False":
            pass
        else:
            raise ValueError("Invalid strandedness value at line no. {} (only True/False accepted): {}".format(
                counter, stranded
            ))
        if len(fields) >= 4:
            score = fields[3]
            if score == "":
                score = 0
            else:
                try:
                    score = float(fields[3])
                except ValueError:
                    raise ValueError("Invalid score specified for {} at line no. {} (only int/floats accepted): {}".format(
                        gff_name, counter, fields[3]
                    ))
            configuration["prepare"]["files"]["source_score"][label] = score
        else:
            configuration["prepare"]["files"]["source_score"][label] = 0

        for arr, pos, default in [("reference", 4, False), ("exclude_redundant", 5, False),
                                  ("strip_cds", 6, False), ("skip_split", 7, False)]:
            try:
                val = fields[pos]
                if val.lower() in ("false", "true"):
                    val = eval(val.capitalize())
                else:
                    raise ValueError(
                        "Malformed line (no. {}). The last fields should be either True or False. At line "
                        "no. {}, the {} value (position {}) is instead {}.".format(counter, counter,
                                                                                   arr, pos, val))
            except IndexError:
                val = default
            if arr == "skip_split":
                configuration["pick"]["chimera_split"]["skip"].append(val)
            else:
                configuration["prepare"]["files"][arr].append(val)

    assert "exclude_redundant" in configuration["prepare"]["files"]

    cfg.prepare.files.gff = configuration["prepare"]["files"]["gff"]
    cfg.prepare.files.labels = configuration["prepare"]["files"]["labels"]
    cfg.prepare.files.strand_specific_assemblies = configuration["prepare"]["files"]["strand_specific_assemblies"]
    cfg.prepare.files.source_score = configuration["prepare"]["files"]["source_score"]
    cfg.prepare.files.reference = configuration["prepare"]["files"]["reference"]
    cfg.prepare.files.exclude_redundant = configuration["prepare"]["files"]["exclude_redundant"]
    cfg.prepare.files.strip_cds = configuration["prepare"]["files"]["strip_cds"]
    cfg.pick.chimera_split.skip = configuration["pick"]["chimera_split"]["skip"]

    return cfg



# coding: utf_8

"""
Generic parser for GTF files.
"""

from .parser import Parser
from .gfannotation import GFAnnotation
import re
from ..exceptions import InvalidParsingFormat


class GtfLine(GFAnnotation):
    """This class defines a typical GTF line, with some added functionality
    to make it useful in e.g. parsing cufflinks GTF files or
    creating GTF lines from scratch.
    Fields:
    - chrom
    - source
    - feature
    - start,end
    - score
    - phase
    - strand
    - info (a dictionary containing all the annotations contained in the last field)

    - gene: the gene_id
    - transcript: the transcript_id

    For cufflinks files, also:
    - nearest_ref
    - tss_id
    - ccode"""

    # _slots=['chrom','source','feature','start',\
    # 'end','score','strand','phase','info']

    _attribute_pattern = re.compile(r"([^;\s]*) \"([^\"]*)\"(?:;|$)")

    def __init__(self, line, my_line='', header=False):

        self.__is_transcript = False
        super().__init__(line, my_line, header=header)
        self.__is_derived, self.__derived_from = None, False  # Placeholders
        self.__is_gene = self._set_is_gene()
        if not self.is_gene:
            self.__is_transcript = self._set_is_transcript()
            self.__set_parent()

    def _parse_attributes(self):
        """
        Method to retrieve the attributes from the last field
        of the GTF line.
        :return:
        """

        attributes = dict((key, self._attribute_definition(val))
                          for key, val in self._attribute_pattern.findall(self._attr.rstrip()))
        self.attributes.update(attributes)
        self.__name = self.__set_name()
        self.__set_gene()
        if "gene_id" not in self.attributes:
            raise ValueError("Missing 'gene_id' in the line")
        if self.gene is None:
            raise TypeError("Invalid null gene_id; attributes: {}".format(self.attributes))

    def _format_attributes(self):

        """
        Private method to format the last field of the GTF line
        prior to printing.
        :return:
        """

        return self._format_attributes_dict(
            self.attributes,
            gene=self.gene,
            transcript=self.transcript)

    @staticmethod
    def _format_attributes_dict(attributes, gene=None, transcript=None):
        """Method to define how to get the attribute string."""

        if not gene:
            pass
        else:
            attributes['gene_id'] = gene
        if "gene_id" in attributes and isinstance(attributes["gene_id"], list):
            attributes["gene_id"] = ",".join(attributes["gene_id"])

        if not transcript:
            attributes["transcript_id"] = attributes.pop("transcript_id", None)
        else:
            attributes["transcript_id"] = transcript

        order = ['gene_id', 'transcript_id', 'exon_number', 'gene_name', 'transcript_name']

        info_list = []

        for tag in order:
            if tag in attributes:
                if isinstance(attributes[tag], list):
                    val = ",".join(attributes[tag])
                else:
                    val = attributes[tag]
                info_list.append("{0} \"{1}\"".format(tag, val))

        done = set()
        for info in iter(key for key in attributes if
                         attributes[key] not in (None, "", []) and key not in order):
            if info in ("Parent", "ID", "parent", "id"):
                continue

            if isinstance(attributes[info], list):
                val = ",".join(attributes[info])
            else:
                val = attributes[info]
            if info == "name":
                info = "Name"
            if info in done:
                continue
            done.add(info)
            info_list.append("{0} \"{1}\"".format(info, val))
        attrs = "; ".join(info_list) + ";"
        return attrs

    @property
    def name(self):
        """
        Returns the name of the feature. It defaults to the ID if missing.
        :rtype str
        """
        return self.__name

    def __set_name(self):

        if "Name" not in self.attributes:
            return self.id
        return self.attributes["Name"]

    @name.setter
    def name(self, *args):
        """
        Setter for name. The argument must be a string.
        :param args:
        :type args: list[(str)] | str
        """

        if not isinstance(args[0], (type(None), str)):
            raise TypeError("Invalid value for name: {0}".format(args[0]))
        self.attributes["Name"] = args[0]
        self.__name = args[0]

    @property
    def is_transcript(self):
        """
        Flag. True if feature is "transcript" or contains "RNA", False in all other cases.
        :rtype : bool
        """
        try:
            return self.__is_transcript
        except AttributeError:
            self.__is_transcript = self._set_is_transcript()
            return self.__is_transcript

    transcript_pattern = re.compile("(^transcript$|rna$)", re.IGNORECASE)

    def _set_is_transcript(self):

        return (self.feature is not None) and (self.transcript_pattern.search(self.feature) is not None)

    def __set_transcript(self):
        if self.header is True:
            self._transcript = None
        else:
            try:
                self._transcript = str(self.attributes["transcript_id"])
            except KeyError:
                raise KeyError(self.attributes)

    @property
    def parent(self):
        """This property looks up the "Parent" field in the "attributes" dictionary.
        If the line is a transcript line, it returns the gene field.
        Otherwise, it returns the transcript field.
        In order to maintain interface consistency with
        the GFF objects and contrary to other attributes,
        this property returns a *list*, not a string. This is due
        to the fact that GFF files support
        multiple inheritance by separating the parent entries with a comma.

        :rtype : list

        """

        return self.__parent

    def __set_parent(self):

        if self.is_gene:
            self.__parent = None
        self.__set_transcript()
        if self.is_transcript is True and self.gene is not None:
            self.__parent = [self.gene]
        elif self.is_transcript is True:
            raise ValueError("No gene")
        else:
            if self.transcript is not None:
                self.__parent = [self.transcript]

    @parent.setter
    def parent(self, parent):
        """
        Setter for parent. Acceptable values
        :param parent: the new parent value
        :type parent: list | str
        """
        if isinstance(parent, str):
            parent = parent.split(",")
        elif isinstance(parent, (int, float)):
            parent = [str(parent)]
        elif isinstance(parent, (bytes,)):
            parent = [parent.decode()]

        self.attributes["Parent"] = parent
        self.__parent = parent
        if self.is_transcript is True:
            self.gene = parent

    # pylint: disable=invalid-name
    @property
    def id(self):
        """
        ID of the line. "transcript_id" for transcript features, None in all other cases.
        :rtype : str | None
        """
        if self.is_transcript is True:
            return self.transcript
        else:
            return None

    @id.setter
    def id(self, newid):
        """
        Setter for id. Only transcript features can have their ID set.
        :param newid: the new ID
        """
        if self.is_transcript is True:
            self.transcript = newid
        else:
            pass
    # pylint: enable=invalid-name

    @property
    def gene(self):
        """
        Return the "gene_id" field.
        :rtype : str | None
        """

        # if "gene_id" not in self.attributes and self.is_transcript is True:
        #     self.attributes["gene_id"] = self.parent[0]
        if self.__gene is None:
            self.__set_gene()
        return self.__gene

    def __set_gene(self):
        try:
            self.__gene = str(self.attributes["gene_id"])
            if self.is_transcript:
                self.__parent = [self.__gene]
        except KeyError:
            pass

    @gene.setter
    def gene(self, gene):
        """
        Setter for gene.
        :param gene:
        :rtype : str

        """

        self.attributes["gene_id"] = self.__gene = gene
        if self.is_transcript:
            self.__parent = [gene]

    @property
    def transcript(self):
        """
        This property returns the "transcript_id" field of the GTF line.
        :rtype : str
        """
        return self._transcript

    @transcript.setter
    def transcript(self, transcript):
        """
        Setter for the transcript attribute. It also modifies the "transcript_id" field.
        :param transcript:
        :type transcript: str
        """

        self.attributes["transcript_id"] = self._transcript = str(transcript)
        if self.is_transcript is True:
            self.attributes["ID"] = transcript
        else:
            self.parent = [transcript]
        self._transcript = transcript

    @property
    def is_parent(self):
        """
        True if we are looking at a transcript, False otherwise.
        :rtype : bool
        """
        if any([self.is_transcript, self.is_gene]):
            return True
        return False

    @property
    def is_derived(self):
        """
        Property. It checks whether there is a "Derives_from" attribute among the line attributes.
        :rtype bool
        """
        if self.__is_derived is None:
            self.__is_derived = self.__set_is_derived()

        return self.__is_derived

    derived_pattern = re.compile("^derives_from$", flags=re.IGNORECASE)

    def __set_is_derived(self):
        return any(self.derived_pattern.search(key) is not None for key in self.attributes)

    @property
    def derived_from(self):
        """
        Boolean property. True if the GTF line has a "derives_from" tag,
        False otherwise.
        """
        if self.__derived_from is False:
            return self.__derived_from

    def __set_derived_from(self):

        if self.is_derived is False:
            return None
        else:
            for key in self.attributes:
                if re.search(self.derived_pattern, key) is not None:
                    return self.attributes[key].split(",")

    gene_pattern = re.compile("gene", re.IGNORECASE)

    @property
    def _negative_order(self):
        return ["3UTR",
                "exon",
                "stop_codon",
                "CDS",
                "start_codon",
                "5UTR"]

    @property
    def _positive_order(self):
        return ["5UTR",
                "exon",
                "start_codon",
                "CDS",
                "stop_codon",
                "3UTR"]


class GTF(Parser):
    """The parsing class."""

    __annot_type__ = "gtf"

    def __init__(self, handle):
        """
        Constructor for the parser.
        :param handle: either the filename or the handle for the file to parse.
        :return:
        """

        super().__init__(handle)
        self.__line_counter = 0

    def __next__(self):
        line = None
        try:
            line = next(self._handle)
            self.__line_counter += 1
            return GtfLine(line)
        except (ValueError, KeyError, TypeError, UnicodeError, AttributeError, AssertionError) as exc:
            error = "Invalid line for file {name}, line {counter}:\n{line}Error: {exc}\n".format(
                name=self.name, counter=self.__line_counter, line=line, exc=exc)
            raise InvalidParsingFormat(error)

    @property
    def file_format(self):
        return self.__annot_type__



"""
Module which contains all functions related to logging.
"""

import logging
import logging.handlers
from dataclasses import field

import argparse

import os
from marshmallow import validate
from marshmallow_dataclass import dataclass, Optional

__author__ = 'Luca Venturini'


@dataclass
class LoggingConfiguration:
    log_level: str = field(default="INFO", metadata={
                "metadata": {
            "description": "Verbosity for SQL calls. Default: WARNING. In decreasing order: 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'",
            }, "validate": validate.OneOf(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    })
    sql_level: str = field(default="WARNING", metadata={
                "metadata": {
            "description": "General verbosity. Default: INFO. In decreasing order: 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'",
            }, "validate": validate.OneOf(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    })
    log: Optional[str] = field(default=None, metadata={
            })


formatter = logging.Formatter(
    "{asctime} - {name} - {filename}:{lineno} - {levelname} - {funcName} \
- {processName} - {message}",
    style="{"
)

null_logger = logging.getLogger("null")
null_handler = logging.NullHandler()
null_handler.setFormatter(formatter)
null_logger.setLevel(logging.CRITICAL)
null_logger.addHandler(null_handler)


def create_null_logger(*args, **kwargs):
    """Static method to create a default logging instance for the loci.
    The default is a null handler (no log).

    :param instance: the instance used to derive a name for the logger. It must be either a string
    or a class instance with a __name__ attribute."""

    if len(args) > 0:
        null_special_logger = logging.getLogger(args[0])
        null_special_handler = logging.NullHandler()
        null_special_handler.setFormatter(formatter)
        null_special_logger.handlers = [null_special_handler]
        if "level" in kwargs:
            null_special_logger.setLevel(kwargs["level"])
        else:
            null_special_logger.setLevel(logging.CRITICAL)
        return null_special_logger

    return null_logger


def check_logger(logger):
    """Quick function to verify that a logger is really a logger,
    otherwise it raises a ValueError.

    :param logger: the logger instance
    :type logger: logging.Logger
    """

    if isinstance(logger, logging.Logger):
        return logger
    else:
        raise ValueError("{0} is not a logger but rather {1}".format(
            logger, type(logger)
        ))


def create_default_logger(name, level="WARN"):
    """Default logger
    :param name: string used to give a name to the logger.
    :type name: str

    :param level: level of the logger. Default: WARN
    """

    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.setLevel(level)
    logger.addHandler(handler)
    logger.propagate = False
    return logger


def create_queue_logger(instance, prefix=""):
    """
    Create a queue logger for a specific class, which *must* have
     a "logging_queue" property redirecting to a queue-like object.
     If the instance possesses a "log_level" attribute, the log level
     will be set to its value; otherwise, the logger will be configured
     with a default level of "WARNING".

    :param instance:
    :param prefix:
    :return:
    """

    if not hasattr(instance, "logging_queue"):
        raise AttributeError(
            "A queue logger can be built only using a class with the \"logging_queue\" property!")
    instance._log_handler = logging.handlers.QueueHandler(instance.logging_queue)
    if hasattr(instance, "log_level"):
        instance._log_handler.setLevel(instance.log_level)
    else:
        instance._log_handler.setLevel("WARNING")

    if prefix:
        instance.logger = logging.getLogger("{0}.{1}".format(prefix,
                                                             instance.name if
                                                             hasattr(instance, "name") else
                                                             "default"))
    else:
        instance.logger = logging.getLogger(instance.name)
    instance.logger.addHandler(instance._log_handler)
    instance.logger.setLevel(instance._log_handler.level)
    instance.logger.propagate = False
    return


def create_logger_from_conf(conf, name="mikado", mode="w"):
    logger = logging.getLogger(name)
    # Remove all previous handlers, if any is present. THIS IS NECESSARY TO PREVENT CLASHES IN THE UNIT-TESTS
    for handler in logger.handlers[:]:
        handler.close()
        logger.removeHandler(handler)

    if isinstance(conf, LoggingConfiguration):
        log_settings = conf
    else:
        log_settings = conf.log_settings
    assert isinstance(log_settings, LoggingConfiguration)
    handle = log_settings.log

    if handle is None:
        handler = logging.StreamHandler()
    else:
        os.makedirs(os.path.dirname(os.path.abspath(log_settings.log)), exist_ok=True)
        handler = logging.FileHandler(log_settings.log, mode=mode)

    handler.setFormatter(formatter)
    logger.setLevel(log_settings.log_level)
    logger.addHandler(handler)
    logger.propagate = False
    return logger



#!/usr/bin/env python3

import argparse
import collections
import sys
import tabulate
import pandas as pd
import re

__doc__ = "Script to collapse various stat files into one."


# 1 Command line:
# 2 /home/lucve/miniconda3/envs/mikado2/bin/mikado compare -r reference.gff3 -p Daijin/5-mikado/pick/permissive/mikado-permissive.loci.gff3 -o compare -l compare.log
# 3 18 reference RNAs in 12 genes
# 4 22 predicted RNAs in  15 genes
# 5 --------------------------------- |   Sn |   Pr |   F1 |
# 6                         Base level: 94.90  83.22  88.68
# 7             Exon level (stringent): 80.56  71.60  75.82
# 8               Exon level (lenient): 91.18  76.54  83.22
# 9                  Splice site level: 95.19  81.15  87.61
# 10                       Intron level: 96.84  88.19  92.31
# 11                  Intron level (NR): 94.34  79.37  86.21
# 12                 Intron chain level: 69.23  50.00  58.06
# 13            Intron chain level (NR): 69.23  50.00  58.06
# 14       Transcript level (stringent): 55.56  45.45  50.00
# 15   Transcript level (>=95% base F1): 72.22  59.09  65.00
# 16   Transcript level (>=80% base F1): 72.22  59.09  65.00
# 17          Gene level (100% base F1): 75.00  60.00  66.67
# 18         Gene level (>=95% base F1): 83.33  66.67  74.07
# 19         Gene level (>=80% base F1): 83.33  66.67  74.07
# 20
# 21 #   Matching: in prediction; matched: in reference.
# 22
# 23             Matching intron chains: 9
# 24              Matched intron chains: 9
# 25    Matching monoexonic transcripts: 4
# 26     Matched monoexonic transcripts: 4
# 27         Total matching transcripts: 13
# 28          Total matched transcripts: 13
# 29
# 30           Missed exons (stringent): 14/72  (19.44%)
# 31            Novel exons (stringent): 23/81  (28.40%)
# 32             Missed exons (lenient): 6/68  (8.82%)
# 33              Novel exons (lenient): 19/81  (23.46%)
# 34                     Missed introns: 3/53  (5.66%)
# 35                      Novel introns: 13/63  (20.63%)
# 36
# 37        Missed transcripts (0% nF1): 0/18  (0.00%)
# 38         Novel transcripts (0% nF1): 3/22  (13.64%)
# 39              Missed genes (0% nF1): 0/12  (0.00%)
# 40               Novel genes (0% nF1): 3/15  (20.00%)

digit = r"[0-9]*\.?[0-9]+"
accuracy_pat = re.compile(r"^\s+(.*):\s+({digit})\s+({digit})\s+({digit})\s+$".format(digit=digit))
match_pat = re.compile(r"\s+(.*): (\d+)")
missed_novel_pat = re.compile(r"\s+(.*): (\d+)/(\d+)\s+\(({digit})%\)".format(digit=digit))


accuracy_stats = {
    6: "Base level", 7: "Exon level (stringent)", 8: "Exon level (lenient)", 9: "Splice site level",
    10: "Intron level", 11: "Intron level (NR)", 12: "Intron chain level", 13: "Intron chain level (NR)",
    14: "Transcript level (stringent)", 15: "Transcript level (>=95% base F1)",
    16: "Transcript level (>=80% base F1)", 17: "Gene level (100% base F1)", 18: "Gene level (>=95% base F1)",
    19: "Gene level (>=80% base F1)"
}


match_stats = {
    23: "Matching intron chains", 24: "Matched intron chains",
    25: "Matching monoexonic transcripts", 26: "Matched monoexonic transcripts",
    27: "Total matching transcripts", 28: "Total matched transcripts"
}


novel_missed_stats = {
    30: "Missed exons (stringent)", 31: "Novel exons (stringent)", 32: "Missed exons (lenient)",
    33: "Novel exons (lenient)", 34: "Missed introns", 35: "Novel introns",
    37: "Missed transcripts (0% nF1)",  38: "Novel transcripts (0% nF1)",
    39: "Missed genes (0% nF1)", 40: "Novel genes (0% nF1)"
}


def accuracy_retrieval(line: str, requested: str) -> (float, float, float):
    level, sn, pr, f1 = accuracy_pat.search(line).groups()
    assert level == requested, (line, requested)
    sn, pr, f1 = [float(val) for val in (sn, pr, f1)]
    assert 0 <= min([sn, pr, f1]) <= max([sn, pr, f1]) <= 100
    return sn, pr, f1


def match_retrieval(line: str, requested: str) -> int:
    level, match = match_pat.search(line).groups()
    assert requested == level, (line, requested)
    match = int(match)
    assert match >= 0
    return match


def missed_novel_retrieval(line: str, requested: str) -> (int, int, float):

    level, found, maximum, proportion = missed_novel_pat.search(line).groups()
    assert level == requested, (line, requested)
    found, maximum = int(found), int(maximum)
    proportion = float(proportion)
    assert 0 <= proportion <= 100
    proportion = f"{proportion}%"
    return found, maximum, proportion


def launch(args):
    if args.avf is True:
        print("Available formats for this script:")
        print(*tabulate.tabulate_formats, sep="\n")
        sys.exit(0)

    data = {
        "Sn": dict((key, collections.defaultdict(dict)) for counter, key in sorted(accuracy_stats.items())),
        "Pr": dict((key, collections.defaultdict(dict)) for counter, key in sorted(accuracy_stats.items())),
        "F1": dict((key, collections.defaultdict(dict)) for counter, key in sorted(accuracy_stats.items())),
        "Matches": dict((key, collections.defaultdict(dict)) for counter, key in sorted(match_stats.items())),
        "MNR":{
            "Found": dict((key, collections.defaultdict(dict)) for counter, key
                          in sorted(novel_missed_stats.items())),
            "Maximum": dict((key, collections.defaultdict(dict)) for counter, key
                              in sorted(novel_missed_stats.items())),
            "Proportion": dict((key, collections.defaultdict(dict)) for counter, key in
                               sorted(novel_missed_stats.items()))
        }
    }

    for stat in args.stat:
        for line_counter, line in enumerate(stat, start=1):
            if line_counter in accuracy_stats:
                level = accuracy_stats[line_counter]
                sn, pr, f1 = accuracy_retrieval(line, level)
                data["Sn"][level][stat.name] = sn
                data["Pr"][level][stat.name] = pr
                data["F1"][level][stat.name] = f1
            elif line_counter in match_stats:
                level = match_stats[line_counter]
                match = match_retrieval(line, level)
                data["Matches"][level][stat.name] = match
                continue
            elif line_counter in novel_missed_stats:
                level = novel_missed_stats[line_counter]
                found, maximum, proportion = missed_novel_retrieval(line, level)
                if stat.name not in data["MNR"]:
                    data["MNR"][stat.name] = collections.defaultdict(dict)

                data["MNR"][stat.name][level]["Total"] = found
                data["MNR"][stat.name][level]["Out of"] = maximum
                data["MNR"][stat.name][level]["Proportion"] = proportion
                continue
            else:
                continue

    for tablefmt in args.tablefmt:
        if tablefmt in ("tsv", "html", "csv", "rst"):
            ext = tablefmt
        elif "latex" in tablefmt:
            ext = f"{tablefmt}.tex"
        else:
            ext = f"{tablefmt}.txt"

        for key in ["Sn", "Pr", "F1", "Matches"]:
            if not ("all" in args.levels or key.lower() in args.levels):
                continue
            if key == "Sn":
                name = "Sensitivity"
            elif key == "Pr":
                name = "Precision"
            elif key in ("F1", "Matches"):
                name = key[:]
            else:
                continue
            lname = name.lower()
            with open(f"{args.out}.{lname}.{ext}", "wt") as out:
                accuracy = pd.DataFrame.from_dict(dict(**data[key]), orient="index")
                accuracy.index.name = f"{name}"
                print(tabulate.tabulate(accuracy, showindex="always", headers="keys", tablefmt=tablefmt),
                      file=out)

        # Now the last table
        if not ("all" in args.levels or "missed_novel" in args.levels):
            continue
        dfs = []
        for fname, fdata in data["MNR"].items():
            df = pd.DataFrame.from_dict(fdata, orient="index")
            df.columns.names = [fname]
            dfs.append(df)

        df = pd.concat(dfs, keys=[df.columns.names[0] for df in dfs], axis=1)
        df.columns.names = ["File", "Statistic"]
        h = [df.columns.names[0] + "\n" + df.columns.names[1]] + [
            "\n".join([fname, stat]) if num % 3 == 1 else "\n" + stat for num, (fname, stat) in
            enumerate(df.columns.tolist())]
        with open(f"{args.out}.missed_novel.{ext}", "wt") as out:
            print(tabulate.tabulate(df, showindex="always", headers=h, tablefmt=tablefmt), file=out)


def parser():
    parser = argparse.ArgumentParser(__doc__)
    parser.add_argument("-o", "--out", default="mikado_compare_selected_stats",
                        help="Prefix for the output files")
    parser.add_argument("-fmt", "--format", nargs="+", choices=tabulate.tabulate_formats,
                        dest="tablefmt", default=["grid"], metavar="",
                        help="List of formats to print the tables into, separated by a space. Available formats can be "
                             "seen using --available-formats. "
                             "Terminate the list with --. Default: %(default)s")
    parser.add_argument("-l", "--levels", nargs="+", choices=["all", "f1", "sn", "pr", "matches", "missed_novel"],
                        default=["f1", "sn", "pr"],
                        help="Levels to print, separated by a space. Terminate the list with --. Default: %(default)s.")
    parser.add_argument("-avf", "--available-formats", action="store_true", default=False, dest="avf",
                        help="Print out a list of available formats and exit.")
    parser.add_argument("stat", type=argparse.FileType("rt"),
                        nargs="*")
    parser.set_defaults(func=launch)
    return parser


if __name__ == "__main__":
    args = parser()
    launch(args)



#!/usr/bin/env python3

import os
import argparse


class AsmStats:

    file = ""
    genes = 0
    monoexonic_genes = 0
    transcripts = 0
    transcripts_per_gene = 0.0
    transcript_len_mean = 0.0
    monoexonic_transcripts = 0
    exons = 0
    exons_per_transcript = 0.0
    exon_len_mean = 0.0

    def __init__(self):
        self.data = []

    def __str__(self):

        return "\t".join([str(_) for _ in
                          [self.file, self.genes, self.monoexonic_genes, self.transcripts,
                           self.transcripts_per_gene, self.transcript_len_mean,
                           self.monoexonic_transcripts, self.exons, self.exons_per_transcript,
                           self.exon_len_mean]])

    @staticmethod
    def header():

        return "\t".join(["File",
                          "genes",
                          "monoexonic_genes",
                          "transcripts",
                          "transcripts_per_gene",
                          "transcript_len_mean",
                          "monoexonic_transcripts",
                          "exons",
                          "exons_per_transcript",
                          "exon_len_mean"
                          ])


def main():
    parser = argparse.ArgumentParser("Script to collect info from multiple mikado util stats files")
    parser.add_argument("input", nargs='+', help="The list of mikado util stats file to process")
    args = parser.parse_args()

    print(AsmStats.header())

    for stats_file in args.input:
        with open(stats_file) as f:
            s = AsmStats()
            s.file = os.path.split(stats_file)[1]
            f.readline()
            # f.readline()
            # f.readline()
            # f.readline()
            for line in f:
                parts = line.split("\t")
                col2 = float(parts[1].replace(',','').replace("NA","0"))
                col3 = float(parts[2].replace(',','').replace("NA","0.0"))
                if parts[0] == "Number of genes":
                    s.genes = col2
                elif parts[0] == "Number of monoexonic genes":
                    s.monoexonic_genes = col2
                elif parts[0] == "Transcripts per gene":
                    s.transcripts = col2
                    s.transcripts_per_gene = col3
                elif parts[0] == "CDNA lengths":
                    s.transcript_len_mean = col3
                elif parts[0] == "Monoexonic transcripts":
                    s.monoexonic_transcripts = col2
                elif parts[0] == "Exons per transcript":
                    s.exons = col2
                    s.exons_per_transcript = col3
                elif parts[0] == "Exon lengths":
                    s.exon_len_mean = col3
            print(s)

main()



# coding: utf-8

"""This module contains the subprograms launched by the Mikado suite"""

# noinspection PyPep8
from . import configure
# noinspection PyPep8
from . import compare
# noinspection PyPep8
from . import pick
# noinspection PyPep8
from . import prepare
# noinspection PyPep8
from . import serialise
# noinspection PyPep8
from . import util



# coding: utf-8

"""
Unit test for a transcript on the negative strand.
"""

import unittest
import operator
import re
import logging
from .. import parsers, loci, exceptions
from ..utilities.log_utils import create_null_logger
from ..utilities.log_utils import create_default_logger


class TranscriptTesterNegative(unittest.TestCase):

    logger = create_null_logger("null")
    logger.setLevel(logging.WARNING)

    tr_gff = """Chr1    TAIR10    mRNA    5928    8737    .    -    .    ID=AT1G01020.1;Parent=AT1G01020
Chr1    TAIR10    five_prime_UTR    8667    8737    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    8571    8666    .    -    0    Parent=AT1G01020.1;
Chr1    TAIR10    exon    8571    8737    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    8417    8464    .    -    0    Parent=AT1G01020.1;
Chr1    TAIR10    exon    8417    8464    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    8236    8325    .    -    0    Parent=AT1G01020.1;
Chr1    TAIR10    exon    8236    8325    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    7942    7987    .    -    0    Parent=AT1G01020.1;
Chr1    TAIR10    exon    7942    7987    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    7762    7835    .    -    2    Parent=AT1G01020.1;
Chr1    TAIR10    exon    7762    7835    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    7564    7649    .    -    0    Parent=AT1G01020.1;
Chr1    TAIR10    exon    7564    7649    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    7384    7450    .    -    1    Parent=AT1G01020.1;
Chr1    TAIR10    exon    7384    7450    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    7157    7232    .    -    0    Parent=AT1G01020.1;
Chr1    TAIR10    exon    7157    7232    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    CDS    6915    7069    .    -    2    Parent=AT1G01020.1;
Chr1    TAIR10    three_prime_UTR    6437    6914    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    exon    6437    7069    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    three_prime_UTR    5928    6263    .    -    .    Parent=AT1G01020.1
Chr1    TAIR10    exon    5928    6263    .    -    .    Parent=AT1G01020.1"""

    tr_lines = [line for line in tr_gff.split("\n") if line]
    for pos, line in enumerate(tr_lines):
        tr_lines[pos] = re.sub(r"\s+", "\t", line)
        assert len(tr_lines[pos].split("\t")) == 9, line.split("\t")

    tr_gff_lines = [parsers.GFF.GffLine(line) for line in tr_lines]

    for l in tr_gff_lines:
        assert l.header is False

    def setUp(self):
        """Basic creation test."""

        self.tr = loci.Transcript(self.tr_gff_lines[0], logger=self.logger)
        for line in self.tr_gff_lines[1:]:
            self.tr.add_exon(line)
        self.tr.name = self.tr.id
        self.tr.finalize()
        self.tr.logger = self.logger

        self.orf = parsers.bed12.BED12()
        self.orf.chrom = self.tr.id
        self.orf.start = 1
        self.orf.end = self.tr.cdna_length
        self.orf.name = self.tr.id
        self.orf.strand = "+"
        self.orf.score = 0
        self.orf.thick_start = self.tr.selected_start_distance_from_tss + 1
        self.orf.thick_end = self.tr.cdna_length - self.tr.selected_end_distance_from_tes
        self.orf.block_count = 1
        self.orf.blockSize = self.tr.cdna_length
        self.orf.block_starts = [0]
        self.orf.has_start_codon = True
        self.orf.has_stop_codon = True
        self.orf.transcriptomic = True
        self.assertFalse(self.orf.invalid)
        self.assertEqual(len(self.tr), self.tr.end - self.tr.start + 1)

    def test_print(self):

        self.maxDiff = None
        real_printed = """Chr1\tTAIR10\tmRNA\t5928\t8737\t.\t-\t.\tID=AT1G01020.1;Parent=AT1G01020;Name=AT1G01020.1
Chr1\tTAIR10\texon\t5928\t6263\t.\t-\t.\tID=AT1G01020.1.exon1;Parent=AT1G01020.1
Chr1\tTAIR10\tthree_prime_UTR\t5928\t6263\t.\t-\t.\tID=AT1G01020.1.three_prime_UTR1;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t6437\t7069\t.\t-\t.\tID=AT1G01020.1.exon2;Parent=AT1G01020.1
Chr1\tTAIR10\tthree_prime_UTR\t6437\t6914\t.\t-\t.\tID=AT1G01020.1.three_prime_UTR2;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t6915\t7069\t.\t-\t2\tID=AT1G01020.1.CDS1;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t7157\t7232\t.\t-\t0\tID=AT1G01020.1.CDS2;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7157\t7232\t.\t-\t.\tID=AT1G01020.1.exon3;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t7384\t7450\t.\t-\t1\tID=AT1G01020.1.CDS3;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7384\t7450\t.\t-\t.\tID=AT1G01020.1.exon4;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t7564\t7649\t.\t-\t0\tID=AT1G01020.1.CDS4;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7564\t7649\t.\t-\t.\tID=AT1G01020.1.exon5;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t7762\t7835\t.\t-\t2\tID=AT1G01020.1.CDS5;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7762\t7835\t.\t-\t.\tID=AT1G01020.1.exon6;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t7942\t7987\t.\t-\t0\tID=AT1G01020.1.CDS6;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7942\t7987\t.\t-\t.\tID=AT1G01020.1.exon7;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t8236\t8325\t.\t-\t0\tID=AT1G01020.1.CDS7;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t8236\t8325\t.\t-\t.\tID=AT1G01020.1.exon8;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t8417\t8464\t.\t-\t0\tID=AT1G01020.1.CDS8;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t8417\t8464\t.\t-\t.\tID=AT1G01020.1.exon9;Parent=AT1G01020.1
Chr1\tTAIR10\tCDS\t8571\t8666\t.\t-\t0\tID=AT1G01020.1.CDS9;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t8571\t8737\t.\t-\t.\tID=AT1G01020.1.exon10;Parent=AT1G01020.1
Chr1\tTAIR10\tfive_prime_UTR\t8667\t8737\t.\t-\t.\tID=AT1G01020.1.five_prime_UTR1;Parent=AT1G01020.1"""

        rp = set(real_printed.split("\n"))
        fp = set(str(self.tr).split("\n"))

        diff = "\n====\n".join(["\n".join(sorted(list(rp - set.intersection(rp, fp)))),
                               "\n".join(sorted(list(fp - set.intersection(rp, fp))))])

        self.assertEqual(real_printed,
                         str(self.tr),
                         diff)

        g_bed12 = "Chr1	5927	8737	ID=AT1G01020.1;coding=True;phase=0	0	-	6914	8666	0"\
                  "	10	336,633,76,67,86,74,46,90,48,167	0,509,1229,1456,1636,1834,2014,2308,2489,2643\tParent=AT1G01020"
        self.assertEqual(g_bed12, self.tr.format("bed12", transcriptomic=False))

        t_bed12 = "AT1G01020.1	0	1623	ID=AT1G01020.1;coding=True;phase=0	0	+	71	809	0	10	"\
                  "167,48,90,46,74,86,67,76,633,336	0,167,215,305,351,425,511,578,654,1287"
        self.assertEqual(t_bed12, self.tr.format("bed12", transcriptomic=True))

    def test_empty(self):

        """
        Test that the inference of exons is valid.
        :return:
        """

        self.tr.finalized = False
        self.tr.exons = []
        self.tr.finalize()
        self.assertEqual(self.tr.strand, "-")
        self.assertEqual(self.tr.number_internal_orfs, 1)
        self.assertEqual(self.tr.exon_num, 10)
        self.assertEqual(self.tr.exon_num, len(self.tr.exons))
        self.assertEqual(self.tr.start, 5928)
        self.assertEqual(self.tr.end, 8737)
        exons = [(5928, 6263), (6437, 7069), (7157, 7232),
                 (7384, 7450), (7564, 7649), (7762, 7835),
                 (7942, 7987), (8236, 8325), (8417, 8464), (8571, 8737)]
        # exons = [intervaltree.Interval(*exon) for exon in exons]
        self.assertEqual(self.tr.exons,
                         exons,
                         self.tr.exons)
        # self.assertRaises(Mikado.exceptions.InvalidTranscript, self.tr.finalize)

    def test_invalid_utr(self):

        """
        Test that a transcript with UTR but no CDS defined will raise an exception.
        :return:
        """

        self.tr.combined_cds = []
        self.tr.finalized = False
        self.assertRaises(exceptions.InvalidTranscript, self.tr.finalize)

    def test_basics(self):

        """
        Test basic assertions about the transcript:

        - chromosome (.chrom) should be Chr1
        - strand should be -
        - number of internal orfs should be 1
        - number of exons should be 10
        - the metric "exon_num" should be 10 as well
        - start should be 5928 (1-based offset)
        - end should be 8737
        - the exons should correspond to those in the original strings (defined here in the list)
          and all of them should be of the "Interval" class

        :return:
        """

        self.assertEqual(self.tr.chrom, "Chr1")
        self.assertEqual(self.tr.strand, "-")
        self.assertEqual(self.tr.number_internal_orfs, 1)
        self.assertEqual(self.tr.exon_num, 10)
        self.assertEqual(self.tr.exon_num, len(self.tr.exons))
        self.assertEqual(self.tr.start, 5928)
        self.assertEqual(self.tr.end, 8737)
        exons = [(5928, 6263), (6437, 7069), (7157, 7232),
                 (7384, 7450), (7564, 7649), (7762, 7835),
                 (7942, 7987), (8236, 8325), (8417, 8464), (8571, 8737)]
        # exons = [intervaltree.Interval(*exon) for exon in exons]
        self.assertEqual(self.tr.exons,
                         exons,
                         self.tr.exons)

    def test_cds(self):
        self.assertEqual(sorted(self.tr.combined_cds),
                         sorted(self.tr.selected_cds))
        cds = [(6915, 7069), (7157, 7232), (7384, 7450), (7564, 7649), (7762, 7835), (7942, 7987),
               (8236, 8325), (8417, 8464), (8571, 8666)]

        self.assertEqual(self.tr.combined_cds,
                         cds,
                         self.tr.combined_cds)
        self.assertEqual(self.tr.selected_cds_start, 8666)
        self.assertEqual(self.tr.selected_cds_end, 6915)

    def test_utr(self):
        self.assertEqual(self.tr.five_utr, [(8667, 8737)])
        self.assertEqual(self.tr.three_utr, [(5928, 6263),
                                             (6437, 6914)])

    def test_utr_metrics(self):

        """Test for UTR exon num, start distance, etc."""

        self.assertEqual(self.tr.five_utr_num, 1)
        self.assertEqual(self.tr.five_utr_num_complete, 0)
        self.assertEqual(self.tr.three_utr_num, 2)
        self.assertEqual(self.tr.three_utr_num_complete, 1)

        self.assertEqual(self.tr.five_utr_length, 8737 + 1 - 8667)
        self.assertEqual(self.tr.three_utr_length, 6263 + 1 - 5928 + 6914 + 1 - 6437)
        self.assertEqual(self.tr.selected_start_distance_from_tss,
                         8738 - 8667,
                         self.tr.selected_end_distance_from_tes)
        self.assertEqual(self.tr.selected_end_distance_from_tes,
                         6263 + 1 - 5928 + 6915 - 6437,
                         self.tr.selected_end_distance_from_tes)
        self.assertEqual(self.tr.selected_end_distance_from_junction,
                         6915 - 6437,
                         self.tr.selected_cds_end)
        self.assertEqual(self.tr.end_distance_from_junction,
                         self.tr.selected_end_distance_from_junction)

    def test_introns(self):

        introns = {(8465, 8570), (8326, 8416), (7988, 8235),
                   (7836, 7941), (7650, 7761), (7451, 7563),
                   (7233, 7383), (7070, 7156), (6264, 6436)}

        self.assertEqual(self.tr.introns,
                         introns,
                         self.tr.introns)

        cds_introns = {(8465, 8570), (8326, 8416), (7988, 8235),
                       (7836, 7941), (7650, 7761), (7451, 7563),
                       (7233, 7383), (7070, 7156)}

        self.assertEqual(self.tr.combined_cds_introns,
                         cds_introns,
                         self.tr.combined_cds_introns)

        selected_cds_introns = {(8465, 8570), (8326, 8416), (7988, 8235),
                                (7836, 7941), (7650, 7761), (7451, 7563),
                                (7233, 7383), (7070, 7156)}

        self.assertEqual(self.tr.selected_cds_introns,
                         selected_cds_introns,
                         self.tr.selected_cds_introns)

    # @unittest.SkipTest
    def test_strip_cds(self):
        """
        Test the "stip_cds" function which (as the name implies) removes completely the CDS
        from a transcript.
        :return:
        """

        with self.assertLogs("null", level="DEBUG") as log_split:
            self.tr.strip_cds()
        self.assertIn("DEBUG:null:Stripping CDS from AT1G01020.1", log_split.output)

        self.assertEqual(self.tr.selected_cds_length, 0)
        self.assertEqual(self.tr.three_utr, [])
        self.assertEqual(self.tr.five_utr, [])
        self.assertEqual(self.tr.selected_cds, [])
        self.assertEqual(self.tr.selected_cds_start, None)
        self.assertEqual(self.tr.selected_cds_end, None)

    def test_remove_utr(self):
        """Test for CDS stripping. We remove the UTRs and verify that start/end
        have moved, no UTR is present, etc."""

        # tr = deepcopy(self.tr)
        self.tr.remove_utrs()

        # tr = deepcopy(self.tr)
        # tr.remove_utrs()
        self.assertEqual(self.tr.selected_cds_start, self.tr.end,
                         ((self.tr.selected_cds_start, self.tr.selected_cds_end),
                          (self.tr.start, self.tr.end)))
        self.assertEqual(self.tr.selected_cds_end, self.tr.start)
        self.assertEqual(self.tr.three_utr, [])
        self.assertEqual(self.tr.five_utr, [])
        combined_cds = [(6915, 7069),
                        (7157, 7232),
                        (7384, 7450),
                        (7564, 7649),
                        (7762, 7835),
                        (7942, 7987),
                        (8236, 8325),
                        (8417, 8464),
                        (8571, 8666)]
        self.assertEqual(self.tr.combined_cds,
                         combined_cds,
                         self.tr.combined_cds)
        self.assertEqual(self.tr.combined_utr, [], self.tr.combined_utr)

    def test_load_orf(self):

        """Test for loading a single ORF. We strip the CDS and reload it."""

        self.tr.strip_cds()
        self.tr.load_orfs([self.orf])

        combined_cds = [(6915, 7069), (7157, 7232), (7384, 7450),
                        (7564, 7649), (7762, 7835), (7942, 7987),
                        (8236, 8325), (8417, 8464), (8571, 8666)]

        self.assertEqual(self.tr.combined_cds,
                         combined_cds,
                         self.tr.combined_cds)
        self.assertEqual(self.tr.selected_cds_start, 8666)
        self.assertEqual(self.tr.selected_cds_end, 6915)

    def test_negative_orf(self):
        """Test loading a negative strand ORF onto a multiexonic transcript.
        This should have no effect.
        """

        self.orf.strand = "-"
        self.tr.strip_cds()
        self.tr.load_orfs([self.orf])
        self.assertEqual(self.tr.selected_cds_start, None)

    def testSegments(self):

        self.assertEqual(self.tr.combined_cds_num, 9)
        self.assertEqual(self.tr.selected_cds_num, 9)
        self.assertEqual(self.tr.highest_cds_exon_number, 9)
        self.assertEqual(self.tr.max_intron_length, 248)
        self.assertEqual(self.tr.number_internal_orfs, 1)

    def test_lengths(self):

        self.assertEqual(self.tr.cdna_length, 1623)
        self.assertEqual(self.tr.selected_cds_length, 738)
        self.assertAlmostEqual(self.tr.combined_cds_fraction, 738 / 1623, delta=0.01)
        self.assertAlmostEqual(self.tr.selected_cds_fraction, 738 / 1623, delta=0.01)

    def test_print_no_cds(self):

        self.maxDiff = None
        # tr = deepcopy(self.tr)
        # tr.finalize()

        real_printed = """Chr1\tTAIR10\tmRNA\t5928\t8737\t.\t-\t.\tID=AT1G01020.1;Parent=AT1G01020;Name=AT1G01020.1
Chr1\tTAIR10\texon\t5928\t6263\t.\t-\t.\tID=AT1G01020.1.exon1;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t6437\t7069\t.\t-\t.\tID=AT1G01020.1.exon2;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7157\t7232\t.\t-\t.\tID=AT1G01020.1.exon3;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7384\t7450\t.\t-\t.\tID=AT1G01020.1.exon4;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7564\t7649\t.\t-\t.\tID=AT1G01020.1.exon5;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7762\t7835\t.\t-\t.\tID=AT1G01020.1.exon6;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t7942\t7987\t.\t-\t.\tID=AT1G01020.1.exon7;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t8236\t8325\t.\t-\t.\tID=AT1G01020.1.exon8;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t8417\t8464\t.\t-\t.\tID=AT1G01020.1.exon9;Parent=AT1G01020.1
Chr1\tTAIR10\texon\t8571\t8737\t.\t-\t.\tID=AT1G01020.1.exon10;Parent=AT1G01020.1"""

        self.assertEqual(real_printed, self.tr.format("gff", with_cds=False))

        real_printed_gtf = """Chr1\tTAIR10\tmRNA\t5928\t8737\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1"; Name "AT1G01020.1";
Chr1\tTAIR10\texon\t5928\t6263\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t6437\t7069\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t7157\t7232\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t7384\t7450\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t7564\t7649\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t7762\t7835\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t7942\t7987\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t8236\t8325\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t8417\t8464\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";
Chr1\tTAIR10\texon\t8571\t8737\t.\t-\t.\tgene_id "AT1G01020"; transcript_id "AT1G01020.1";"""

        self.assertEqual(real_printed_gtf, self.tr.__str__(print_cds=False, to_gtf=True))

    def testDoubleOrf(self):

        """Test to verify the introduction of multiple ORFs."""

        self.tr.strip_cds()
        self.tr.finalized = False

        first_orf = parsers.bed12.BED12()
        first_orf.chrom = self.tr.id
        first_orf.start = 1
        first_orf.end = self.tr.cdna_length
        first_orf.name = self.tr.id
        first_orf.strand = "+"
        first_orf.score = 0
        first_orf.thick_start = 100
        first_orf.thick_end = 501
        first_orf.block_count = 1
        first_orf.blockSize = self.tr.cdna_length
        first_orf.block_starts = [0]
        first_orf.has_start_codon = True
        first_orf.has_stop_codon = True
        first_orf.transcriptomic = True
        self.assertFalse(first_orf.invalid, (len(first_orf), first_orf.cds_len))

        # This should not be incorporated
        second_orf = parsers.bed12.BED12()
        second_orf.chrom = self.tr.id
        second_orf.start = 0
        second_orf.end = self.tr.cdna_length
        second_orf.name = "second"
        second_orf.strand = "+"
        second_orf.score = 1
        second_orf.thick_start = 300
        second_orf.thick_end = 401
        second_orf.block_count = 1
        second_orf.blockSize = self.tr.cdna_length
        second_orf.block_starts = [0]
        second_orf.has_start_codon = True
        second_orf.has_stop_codon = True
        second_orf.transcriptomic = True
        self.assertFalse(second_orf.invalid, (len(second_orf), second_orf.cds_len))

        self.assertTrue(loci.Transcript.is_overlapping_cds(first_orf, second_orf))

        # This should be added
        third_orf = parsers.bed12.BED12()
        third_orf.chrom = self.tr.id
        third_orf.start = 1
        third_orf.end = self.tr.cdna_length
        third_orf.name = "third"
        third_orf.strand = "+"
        third_orf.score = 0
        third_orf.thick_start = 1000
        third_orf.thick_end = 1602
        third_orf.block_count = 1
        third_orf.blockSize = self.tr.cdna_length
        third_orf.block_starts = [0]
        third_orf.has_start_codon = True
        third_orf.has_stop_codon = True
        third_orf.transcriptomic = True
        self.assertFalse(third_orf.invalid, (len(third_orf), third_orf.cds_len))

        self.assertFalse(
            loci.Transcript.is_overlapping_cds(
                first_orf, third_orf))
        self.assertFalse(
            loci.Transcript.is_overlapping_cds(
                second_orf, third_orf))

        self.assertFalse(third_orf == second_orf)
        self.assertFalse(first_orf == second_orf)
        self.assertFalse(first_orf == third_orf)

        candidates = [first_orf, second_orf, third_orf]
        self.tr.load_orfs(candidates)

        self.assertTrue(self.tr.is_complete)
        self.tr.finalize()
        self.assertEqual(self.tr.start, 5928)

        self.assertEqual(self.tr.number_internal_orfs, 2, "\n".join(
            [str(x) for x in self.tr.internal_orfs]))

        self.assertEqual(self.tr.combined_cds_length, 1005)
        self.assertEqual(self.tr.selected_cds_length, 603)

        new_transcripts = sorted(self.tr.split_by_cds(), key=operator.attrgetter("start"))
        self.assertEqual(len(new_transcripts), 1)
        self.tr.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"
        new_transcripts = sorted(self.tr.split_by_cds(), key=operator.attrgetter("start"))

        self.assertEqual(len(new_transcripts), 2)
        self.assertEqual(new_transcripts[0].five_utr_length, 0)
        self.assertNotEqual(new_transcripts[0].three_utr_length, 0)
        self.assertEqual(new_transcripts[0].cdna_length, 624, msg="{0}-{1}{2}".format(
            new_transcripts[0].start,
            new_transcripts[0].end,
            new_transcripts[0].strand,
        ))
        self.assertEqual(new_transcripts[0].start, self.tr.start)
        self.assertEqual(new_transcripts[0].end, 6724)

        self.assertEqual(new_transcripts[1].three_utr_length, 0)
        self.assertEqual(new_transcripts[1].end, 8737)


class WheatTest(unittest.TestCase):

    def setUp(self):

        trlines= """Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS\tta_complete\ttranscript\t217224\t221661\t100\t-\t.\tgene_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; transcript_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; Name "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; canonical_junctions "1,2,3,4,5"; canonical_proportion "1.0"; canonical_number "5"; target "Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4 1 1251 +";
Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS\tta_complete\texon\t217224\t218053\t.\t-\t.\tgene_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; transcript_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1";
Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS\tta_complete\texon\t218145\t218384\t.\t-\t.\tgene_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; transcript_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1";
Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS\tta_complete\texon\t218474\t219472\t.\t-\t.\tgene_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; transcript_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1";
Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS\tta_complete\texon\t219557\t220122\t.\t-\t.\tgene_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; transcript_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1";
Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS\tta_complete\texon\t220205\t220327\t.\t-\t.\tgene_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; transcript_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1";
Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS\tta_complete\texon\t220411\t221661\t.\t-\t.\tgene_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1"; transcript_id "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1";"""

        trlines = [parsers.GTF.GtfLine(_) for _ in trlines.split("\n")]

        self.tr = loci.Transcript(trlines[0])
        for _ in trlines[1:]:
            self.tr.add_exon(_)

        self.bed1 = parsers.bed12.BED12()
        self.bed1.header = False
        self.bed1.chrom = self.tr.id
        self.bed1.transcriptomic = True
        self.bed1.start = 1
        self.bed1.end = 4009
        self.bed1.name = "g.4208_TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1|m.4208_type:complete_len:193_(+)"
        self.bed1.strand = "+"
        self.bed1.thick_start = 297
        self.bed1.thick_end = 875
        self.bed1.has_start_codon = True
        self.bed1.has_stop_codon = True
        self.bed1.score = 0
        self.assertEqual(self.bed1.cds_len, 579)

        self.bed2 = parsers.bed12.BED12()
        self.bed2.header = False
        self.bed2.chrom = self.tr.id
        self.bed2.transcriptomic = True
        self.bed2.start = 1
        self.bed2.end = 4009
        self.bed2.name = "TGAC_Trinity_Triticum_aestivum_CS42_TGACv1_scaffold_018984_1AS:216942_226036:c0_g1_i4.mrna1|m.4209_type:complete_len:116_(+)"
        self.bed2.strand = "+"
        self.bed2.thick_start = 2886
        self.bed2.thick_end = 3233
        self.bed2.has_start_codon = True
        self.bed2.has_stop_codon = True
        self.bed2.score = 0
        self.assertEqual(self.bed2.cds_len, 348)

    def test_load(self):
        
        self.tr.load_orfs([self.bed1, self.bed2])
        self.assertEqual(self.tr.number_internal_orfs, 2)
        self.assertEqual(self.tr.selected_cds_start, 221365)
        self.assertEqual(self.tr.selected_cds_end, 220787)

        self.assertEqual(self.tr.combined_cds_start, 221365)
        self.assertEqual(self.tr.combined_cds_end, 218000)

    def test_split(self):

        self.tr.load_orfs([self.bed1, self.bed2])
        self.tr.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"
        new_transcripts = sorted([_ for _ in self.tr.split_by_cds()],
                                 key=operator.attrgetter("start", "end"))
        self.assertEqual(len(new_transcripts), 2, self.tr.configuration.pick.chimera_split)

        self.assertEqual(new_transcripts[0].start, 217224)
        self.assertEqual(new_transcripts[0].end, 218527)

        self.assertEqual(new_transcripts[1].start, 220787)
        self.assertEqual(new_transcripts[1].end, 221661)

        self.assertEqual(new_transcripts[0].selected_cds_start, 218527)
        self.assertEqual(new_transcripts[0].selected_cds_end, 218000)

        self.assertEqual(new_transcripts[1].selected_cds_start, 221365)
        self.assertEqual(new_transcripts[1].selected_cds_end, 220787)


if __name__ == '__main__':
    unittest.main()



"""
This module implements the Hit serialisation class.
"""
from typing import Dict, List
from ...exceptions import InvalidHit
from sqlalchemy import Column, Integer, Float, ForeignKey, Index
from sqlalchemy.sql.schema import PrimaryKeyConstraint
from sqlalchemy.orm import relationship, column_property, backref
from sqlalchemy import select
from sqlalchemy.ext.hybrid import hybrid_property  # hybrid_method
from ...utilities.dbutils import DBBASE
from .query import Query
from .target import Target
from .hsp import Hsp, prepare_hsp
import numpy as np
from ...parsers.blast_utils import merge


__author__ = 'Luca Venturini'


class Hit(DBBASE):
    """This class is used to serialise and store in a DB a BLAST hit.
    Stored attributes:

    - id                Indexing key
    - query_id            Foreign ID key for the query table
    - target_id            Foreign ID key for the target table
    - qt_constrating
    """

    __tablename__ = "hit"
    query_id = Column(Integer, ForeignKey(Query.query_id), unique=False)
    target_id = Column(Integer, ForeignKey(Target.target_id), unique=False)
    evalue = Column(Float)
    bits = Column(Float)
    global_identity = Column(Float)
    global_positives = Column(Float)
    query_start = Column(Integer)
    query_end = Column(Integer)
    target_start = Column(Integer)
    target_end = Column(Integer)
    hit_number = Column(Integer)
    query_multiplier = Column(Float)  # Probably I should move this to a separate table!
    target_multiplier = Column(Float)
    query_aligned_length = Column(Integer)
    target_aligned_length = Column(Integer)

    # query_id, target_id, evalue, bits, global_identity, global_positives, query_start, query_end, target_start,
    # target_end, hit_number, query_multiplier, target_multiplier, query_aligned_length, target_aligned_length

    # Indices and constraints
    qt_constraint = PrimaryKeyConstraint("query_id", "target_id", name="hit_id")
    qt_index = Index("qt_index", "query_id", "target_id", unique=True)
    query_index = Index("hit_query_idx", "query_id", unique=False)
    target_index = Index("hit_target_idx", "target_id", unique=False)
    evalue_index = Index('hit_evalue_idx', 'evalue', unique=False)

    query_object = relationship(Query, uselist=False,
                                lazy="select",
                                innerjoin=True,
                                backref=backref("hits", cascade="all, delete-orphan"))
    target_object = relationship(Target,
                                 uselist=False,
                                 lazy="select",
                                 innerjoin=True,
                                 backref=backref("hits", cascade="all, delete-orphan"))

    @hybrid_property
    def query(self):
        return self.query_object.query_name

    @hybrid_property
    def query_length(self):
        return self.query_object.query_length

    @hybrid_property
    def target(self):
        return self.target_object.target_name

    @hybrid_property
    def target_length(self):
        return self.target_object.target_length

    join_condition = "and_(Hit.query_id==Hsp.query_id, Hit.target_id==Hsp.target_id)"
    hsps = relationship(Hsp, uselist=True,
                        lazy="subquery",
                        backref=backref("hit_object", uselist=False,
                                        overlaps="hits,target_object,query_object"),
                        cascade="all, delete-orphan",
                        single_parent=True,
                        foreign_keys=[query_id, target_id],
                        overlaps="hits,target_object,query_object",
                        primaryjoin=join_condition)

    __table_args__ = (qt_constraint, qt_index, query_index, target_index, evalue_index)

    # All arguments are necessary and it is more convenient to have them here
    # rather than in a struct/dict/whatever
    # pylint: disable=too-many-arguments
    def __init__(self, query_id, target_id,
                 evalue, bits, global_identity, global_positives, query_start, query_end, target_start,
                 target_end, query_aligned_length,
                 target_aligned_length,
                 query_length, alignment,
                 hit_number=1, query_multiplier=1,
                 target_multiplier=1):
        """This function takes as input the id of a target, the id of the query,
        and a hit-object from the XML. The multiplier keyword is used to calculate
        the ratio between the query and the target.

        :param query_id: reference key for the Query table
        :type query_id: int

        :param target_id: reference key for the Target table
        :type target_id: int

        :param alignment: The BLAST Hit object

        :param evalue: Evalue of the hit (recovered from the "description" object)
        :type evalue: float

        :param bits: BitScore of the hit (recovered from the "description" object)
        :type bits: float

        :param hit_number: progressive index that indicates the priority
        of the hit in the database
        :type hit_number: int

        :param query_multiplier: either 1 or 3, it depends on the type of BLAST
        :type query_multiplier: int

        :param target_multiplier: either 1 or 3, it depends on the type of BLAST
        :type target_multiplier: int
        """

        self.query_id, self.query_start, self.query_end = query_id, query_start, query_end
        self.query_aligned_length = query_aligned_length

        self.target_id, self.target_start, self.target_end = target_id, target_start, target_end
        self.target_aligned_length = target_aligned_length

        self.hit_number = hit_number
        self.evalue, self.bits = evalue, bits

        self.global_identity, self.global_positives = global_identity, global_positives
        self.query_multiplier = query_multiplier
        self.target_multiplier = target_multiplier

    # pylint: enable=too-many-arguments

    def __str__(self):
        line = [self.query, self.target, self.evalue,
                self.bits, self.query_start, self.query_end,
                self.target_start, self.target_end,
                self.query_length, self.target_length]

        return "\t".join(str(x) for x in line)

    @classmethod
    def from_hit(cls, query_id, target_id, query_length, alignment, query_multiplier=1, target_multiplier=1,
                 hit_number=1) -> (object, List[Hsp]):
        """"""

        prepared_hit, prepared_hsps = prepare_hit(alignment, query_id, target_id,
                                                  query_multiplier=query_multiplier,
                                                  target_multiplier=target_multiplier,
                                                  query_length=query_length)
        hsps = []
        for counter, hsp in enumerate(prepared_hsps, start=1):
            hsps.append(Hsp.from_dict(counter=counter, query_id=query_id, target_id=target_id,
                                      hsp_dict=hsp))

        new = Hit(query_id=query_id, target_id=target_id, query_multiplier=query_multiplier,
                  hit_number=hit_number, target_multiplier=target_multiplier, **prepared_hit)

        return new, hsps

    @classmethod
    def as_dict_static(cls, state_obj):
        """Method to return a dict representation of the object.
        Static method to be called from outside the class.

        :param state_obj: a namedtuple or an instance of this class

        :rtype: dict
        """

        keys = [
            "evalue",
            "bits",
            "global_identity",
            "global_positives",
            "query_start",
            "query_end",
            "target_start",
            "target_end",
            "hit_number",
            "query_multiplier",
            "target_multiplier",
            "query_aligned_length",
            "target_aligned_length",
        ]

        state = dict().fromkeys(keys)

        for key in keys:
            state[key] = getattr(state_obj, key)

        return state

    @classmethod
    def as_full_dict_static(cls, hit_tuple,
                            hsp_list,
                            query_tuple,
                            target_tuple):
        """
        :param hit_tuple: Hit namedtuple (from direct query to the DB)
        :type hit_tuple: collections.namedtuple
        :param hsp_list: list of hsp dictionaries from Hsp.as_dict_static
        :type hsp_list: list(collections.namedtuple)

        :param query_tuple: Query namedtuple
        :type query_tuple: collections.namedtuple
        :param target_tuple: Target namedtuple
        :type target_tuple: collections.namedtuple
        :rtype: dict
        """

        state = cls.as_dict_static(hit_tuple)
        hsps = [Hsp.as_dict_static(h) for h in hsp_list]

        state["query"] = query_tuple.query_name
        state["target"] = target_tuple.target_name
        state["query_length"] = query_tuple.query_length
        state["target_length"] = target_tuple.target_length
        state["query_hit_ratio"] = (query_tuple.query_length / hit_tuple.query_multiplier) / \
                                   (target_tuple.target_length / hit_tuple.target_multiplier)
        state["hit_query_ratio"] = (target_tuple.target_length / hit_tuple.target_multiplier) / \
                                   (query_tuple.query_length / hit_tuple.query_multiplier)
        state["query_cov"] = state["query_aligned_length"] / query_tuple.query_length
        assert state["query_cov"] <= 1, (state,)
        state["target_cov"] = state["target_aligned_length"] / target_tuple.target_length
        assert state["target_cov"] <= 1, (state,)

        state["hsps"] = []
        for hsp in hsps:
            hsp["query_hsp_cov"] = (hsp["query_hsp_end"] - hsp["query_hsp_start"] + 1)
            hsp["query_hsp_cov"] /= (state["query_length"])
            hsp["target_hsp_cov"] = (hsp["target_hsp_end"] - hsp["target_hsp_start"] + 1)
            hsp["target_hsp_cov"] /= (state["target_length"])
            state["hsps"].append(hsp)

        return state

    def as_dict(self):
        """Method to return a dict representation of the object.
        Necessary for storing.

        :rtype: dict
        """

        state = self.as_dict_static(self)

        # Retrieving the values ONCE
        query_object = self.query_object.as_tuple()
        target_object = self.target_object.as_tuple()

        state["query"] = query_object.query_name
        state["target"] = target_object.target_name
        state["query_length"] = query_object.query_length
        state["target_length"] = target_object.target_length
        state["query_hit_ratio"] = state["query_length"] / state["query_multiplier"] /\
            (state["target_length"] / state["target_multiplier"])

        state["hit_query_ratio"] = state["target_length"] / state["target_multiplier"] /\
            (state["query_length"] / state["query_multiplier"])

        state["hsps"] = []
        for hsp in self.hsps:
            dict_hsp = hsp.as_dict()
            dict_hsp["query_hsp_cov"] = (dict_hsp["query_hsp_end"] -
                                         dict_hsp["query_hsp_start"] + 1)
            dict_hsp["query_hsp_cov"] /= state["query_length"]
            dict_hsp["target_hsp_cov"] = (dict_hsp["target_hsp_end"] -
                                          dict_hsp["target_hsp_start"] + 1)
            dict_hsp["target_hsp_cov"] /= state["target_length"]
            state["hsps"].append(dict_hsp)

        return state

    @hybrid_property
    def query_hit_ratio(self):
        """
        This property returns the quotient (Query Length)/(Target Length)
        """

        ratio = self.query_length / self.query_multiplier
        ratio /= self.target_length / self.target_multiplier

        return ratio

    @hybrid_property
    def hit_query_ratio(self):
        """
        This property returns the quotient (Target Length)/(Query Length)
        """

        ratio = self.target_length / self.target_multiplier
        ratio /= (self.query_length / self.query_multiplier)

        return ratio


hit_cols = [col.name for col in Hit.__table__.columns]
hsp_cols = [col.name for col in Hsp.__table__.columns]


def prepare_hit(hit, query_id, target_id, off_by_one=False, as_list=False,
                **kwargs) -> (Dict, List[Dict]):
    """Prepare the dictionary for fast loading of Hit and Hsp objects.
    global_positives: the similarity rate for the global hit *using the query perspective*
    global_identity: the identity rate for the global hit *using the query perspective*

    :param hit: the hit to parse.
    :type hit: Bio.SearchIO.Hit

    :param query_id: the numeric ID of the query in the database. Necessary for serialisation.
    :type query_id: int

    :param target_id: the numeric ID of the target in the database. Necessary for serialisation.
    :type target_id: int

    :param query_multiplier: the multiplier used for the query (usually either 1 or 3)
    :param target_multiplier: the multiplier used for the query (usually either 1 or 3)

    :param kwargs: additional properties to give to the hit_dict. Retrieved e.g. from descriptions.
    """

    hit_dict = dict()
    hsp_dict_list = []
    q_intervals = []
    t_intervals = []

    qmulti = kwargs["query_multiplier"]
    tmulti = kwargs["target_multiplier"]
    qlength = int(kwargs["query_length"])
    assert isinstance(qmulti, (int, float)), type(qmulti)
    assert isinstance(tmulti, (int, float)), type(tmulti)
    hit_dict.update(kwargs)
    hit_dict["query_id"] = query_id
    hit_dict["target_id"] = target_id

    query_array = np.zeros([2, int(qlength)], dtype=int)

    for counter, hsp in enumerate(hit.hsps):
        if hsp.query_start + off_by_one - 1 > qlength:
            raise ValueError("Invalid length: {}, {}", hsp.query_start + off_by_one - 1, qlength)

        hsp_dict, ident, posit = prepare_hsp(hsp, counter, qmultiplier=qmulti, tmultiplier=tmulti,
                                             off_by_one=off_by_one)
        assert isinstance(ident, np.ndarray)
        assert isinstance(posit, np.ndarray)
        if ident.max() > query_array.shape[1] or ident.min() < 0:
            raise IndexError("Invalid indexing values (max {}; frame {}; hsp: {})!"\
"Too low: {}\nToo high: {}".format(
                query_array.shape[1],
                hsp.query_frame, hsp.__dict__,
                list(ident[(ident < 0)]),
                list(ident[ident > query_array.shape[1]])))
        try:
            query_array[0, ident - off_by_one] = 1
        except IndexError as exc:
            raise IndexError("{}, off by one: {}; min, max: {}, {}; hsp {}".format(
                exc, off_by_one, ident.min(), ident.max(), hsp._items[0].__dict__))
        try:
            query_array[1, posit - off_by_one] = 1
        except IndexError as exc:
            raise IndexError("{}, off by one: {}; min, max: {}, {}; frame: {}".format(
                exc, off_by_one, posit.min(), posit.max(), hsp.query_frame))
        hsp_dict["query_id"] = query_id
        hsp_dict["target_id"] = target_id
        hsp_dict_list.append(hsp_dict)
        q_intervals.append((hsp.query_start, hsp.query_end - 1))
        t_intervals.append((hsp.hit_start, hsp.hit_end - 1))

    q_merged_intervals, q_aligned = merge(q_intervals)
    hit_dict["query_aligned_length"] = min(qlength, q_aligned)
    qstart, qend = q_merged_intervals[0][0], q_merged_intervals[-1][1]
    hit_dict["query_start"], hit_dict["query_end"] = qstart, qend
    identical = np.where(query_array[0] == 1)[0]
    positives = np.where(query_array[1] == 1)[0]

    if identical.shape[0] > q_aligned:
        raise ValueError(
            "Number of identical positions ({}) greater than number of aligned positions ({})!\n{}\n{}".format(
            identical.shape[0], q_aligned, q_intervals, q_merged_intervals))

    if positives.shape[0] > q_aligned:
        raise ValueError("Number of identical positions ({}) greater than number of aligned positions ({})!".format(
            positives.shape[0], q_aligned))

    t_merged_intervals, t_aligned = merge(t_intervals)
    hit_dict["target_aligned_length"] = min(t_aligned, hit.seq_len)
    hit_dict["target_start"] = t_merged_intervals[0][0]
    hit_dict["target_end"] = t_merged_intervals[-1][1] + 1
    hit_dict["global_identity"] = identical.shape[0] * 100 / q_aligned
    hit_dict["global_positives"] = positives.shape[0] * 100 / q_aligned

    if as_list is True:
        hit_list = [hit_dict[col] for col in hit_cols]
        hsp_dict_list = [[hsp[col] for col in hsp_cols] for hsp in hsp_dict_list]
        return hit_list, hsp_dict_list
    else:
        return hit_dict, hsp_dict_list



import rapidjson as json
import yaml
import functools
import toml
import os
from ..configuration import DaijinConfiguration
from . import get_sub_commands, system_hpc_yaml, NOW
import pkg_resources
import shutil
import sys
import inspect
try:
    import drmaa
    _ = drmaa.Session()
    drmaa_available = True
except (RuntimeError, ImportError, AttributeError):
    drmaa_available = False
import snakemake
try:
    from yaml import CSafeLoader as yLoader
except ImportError:
    from yaml import SafeLoader as yLoader


def assemble_transcripts_pipeline(args):

    """
    This section of Daijin is focused on creating the necessary configuration for
    driving the pipeline.
    :param args: CLI arguments from argparse
    :return:
    """
    if args.config.endswith("json"):
        loader = json.load
    elif args.config.endswith("yaml"):
        loader = functools.partial(yaml.load, Loader=yLoader)
    else:
        loader = functools.partial(toml.load)

    with open(args.config, 'r') as _:
        doc = loader(_)

    if args.exe and os.path.exists(args.exe):
        if args.exe.endswith("json"):
            loader = json.load
        else:
            loader = functools.partial(yaml.load, Loader=yLoader)
        with open(args.exe) as _:
            doc["load_cmd"] = loader(_)

    # print(doc["load"])
    # Check the configuration
    _ = DaijinConfiguration.Schema().load(doc)

    # pylint: disable=invalid-name
    if not ("short_reads" in doc or "long_reads" in doc):
        print("No short reads section or long reads sections was present in the configuration. Please include your "
              "samples and try again", file=sys.stderr)
        exit(1)

    LABELS = []
    R1 = []
    R2 = []
    LR_LABELS = []
    LR_FILES = []

    if "short_reads" in doc:
        LABELS = doc["short_reads"]["samples"]
        R1 = doc["short_reads"]["r1"]
        R2 = doc["short_reads"]["r2"]

    if "long_reads" in doc:
        LR_LABELS = doc["long_reads"]["samples"]
        LR_FILES = doc["long_reads"]["files"]
    READS_DIR = doc["out_dir"] + "/1-reads"
    SCHEDULER = doc["scheduler"] if doc["scheduler"] else ""
    CWD = os.path.abspath(".")
    # pylint: enable=invalid-name

    res_cmd, sub_cmd = get_sub_commands(SCHEDULER, args.prefix, args.additional_drmaa)

    # Create log folder
    if not os.path.exists("daijin_logs"):
        os.makedirs("daijin_logs")
    elif not os.path.isdir("daijin_logs"):
        raise OSError("{} is not a directory!".format("daijin_logs"))

    if (len(R1) != len(R2)) and (len(R1) != len(LABELS)):
        print("R1, R2 and LABELS lists are not the same length.  Please check and try again")
        exit(1)

    if len(LR_LABELS) != len(LR_FILES):
        print("long read samples and file arrays in the configuration file are not the same length. Please check and "
              "try again")
        exit(1)

    if not os.path.exists(READS_DIR):
        os.makedirs(READS_DIR)

    for read1, read2, label in zip(R1, R2, LABELS):
        suffix = read1.split(".")[-1]
        if suffix not in ("gz", "bz2"):
            suffix = ""
        else:
            suffix = ".{}".format(suffix)

        r1out = READS_DIR + "/" + label + ".R1.fq{}".format(suffix)
        r2out = READS_DIR + "/" + label + ".R2.fq{}".format(suffix)
        if not os.path.islink(r1out):
            os.symlink(os.path.abspath(read1), r1out)

        if not os.path.islink(r2out):
            os.symlink(os.path.abspath(read2), r2out)

    for lr_file, label in zip(LR_FILES, LR_LABELS):
        suffix = lr_file.split(".")[-1]
        compress = ""
        if suffix in ("gz", "bz2"):
            compress = "." + suffix[:]
            suffix = lr_file.split(".")[-2]

        if suffix in ("fa", "fna", "fasta"):
            suffix = ".fa" + compress
        elif suffix in ("fq", "fastq"):
            suffix = ".fq" + compress
        else:
            suffix = ".{}".format(suffix)

        out = READS_DIR + "/" + label + ".long{}".format(suffix)
        if not os.path.islink(out):
            os.symlink(os.path.abspath(lr_file), out)

    # Launch using SnakeMake
    assert pkg_resources.resource_exists("Mikado", os.path.join("daijin", "assemble.smk"))

    additional_config = {}
    if args.threads is not None:
        additional_config["threads"] = args.threads

    cluster_var = None
    if args.no_drmaa is True and sub_cmd:
        cluster_var = sub_cmd + res_cmd

    drmaa_var = None
    if args.no_drmaa is False and res_cmd:
        if drmaa_available is False:
            print("WARNING: DRMAA not installed or not configured properly. Switching to local/cluster mode. Please "
                  "use the \"-nd\" flag to run Daijin if you do not plan to use DRMAA.", file=sys.stderr)
            drmaa_var = None
            args.no_drmaa = True
        else:
            drmaa_var = res_cmd

    if SCHEDULER == "local":
        hpc_conf = None
        drmaa_var = None
        cluster_var = None
    elif drmaa_var or cluster_var:
        if os.path.exists(args.hpc_conf):
            hpc_conf = args.hpc_conf
        else:
            hpc_conf = system_hpc_yaml
    else:
        hpc_conf = None

    yaml_file = open(os.path.join(doc["out_dir"], "daijin.{}.yaml".format(NOW)), "wt")
    yaml.dump(doc, yaml_file)
    yaml_file.flush()
    shutil.copystat(args.config, yaml_file.name)

    if args.latency_wait is not None:
        latency = abs(args.latency_wait)
    elif SCHEDULER not in ('', "local"):
        latency = 60
    else:
        latency = 1

    kwds = {
        "dryrun": args.dryrun,
        "cores": args.cores,
        "nodes": args.jobs,
        "config": additional_config,
        "workdir": CWD,
        "cluster_config": hpc_conf,
        "cluster": cluster_var,
        "drmaa": drmaa_var,
        "printshellcmds": True,
        "snakemakepath": shutil.which("snakemake"),
        "use_conda": args.use_conda,
        "stats": "daijin_tr_" + NOW + ".stats",
        "force_incomplete": args.rerun_incomplete,
        "detailed_summary": args.detailed_summary,
        "list_resources": args.list,
        "latency_wait": latency,
        "printdag": args.dag,
        "forceall": args.dag,
        "forcerun": args.forcerun,
        "lock": (not args.nolock),
        "printreason": True
    }

    if "configfile" in inspect.getfullargspec(snakemake.snakemake).args:
        kwds["configfile"] = yaml_file.name
    elif "configfiles" in inspect.getfullargspec(snakemake.snakemake).args:
        kwds["configfiles"] = [yaml_file.name]
    else:
        raise KeyError("No configfile key found")

    snakemake.snakemake(
        pkg_resources.resource_filename("Mikado",
                                        os.path.join("daijin", "assemble.smk")),
        **kwds
    )



import argparse
import dataclasses
from typing import Union
import io
from ..configuration import MikadoConfiguration, DaijinConfiguration
import logging
import os

from ..exceptions import InvalidConfiguration
from ..utilities.log_utils import create_logger_from_conf


def _set_pick_mode(conf: Union[MikadoConfiguration, DaijinConfiguration], mode: str) -> Union[
        MikadoConfiguration, DaijinConfiguration]:
    if mode is not None:
        mode = mode.lower()
        if mode == "nosplit":
            conf.pick.chimera_split.execute = False
            conf.pick.chimera_split.blast_check = False
        else:
            conf.pick.chimera_split.execute = True
            if mode == "split":
                conf.pick.chimera_split.blast_check = False
            else:
                conf.pick.chimera_split.blast_check = True
                conf.pick.chimera_split.blast_params.leniency = mode.upper()

    errors = conf.pick.chimera_split.Schema().validate(dataclasses.asdict(conf.pick.chimera_split))
    if len(errors) > 0:  # Dict of errors
        raise InvalidConfiguration(f"Invalid mode selected for Mikado: {mode}\nErrors: {errors}\nDict: "
                                   f"{dataclasses.asdict(conf.pick.chimera_split)}")

    return conf


def check_log_settings_and_create_logger(conf: Union[MikadoConfiguration, DaijinConfiguration],
                                         log: Union[str, io.TextIOWrapper, io.BufferedWriter, bytes, None],
                                         log_level: Union[str, None],
                                         section=None) -> (
        Union[MikadoConfiguration, DaijinConfiguration], logging.Logger):
    """
    Quick method to check the consistency of log settings from the namespace and create a logger.
    :param conf: the configuration object
    :param log: Handle to use for the logger. Use "stderr" for using a streaming logger. If log is None:
    - if the section (see below) if also None, use stderr
    - else, try to infer the log from the parameter in the section
    :param log_level: level of the log to use
    :param section: section of the configuration to use, if any. One between None, pick, serialise, prepare
    :return: MikadoConfiguration, logger
    """

    if not isinstance(conf, MikadoConfiguration):
        raise InvalidConfiguration(f"Invalid object type; it should be MikadoConfiguration, it is {type(conf)}")
    out_dir = getattr(conf, section).files.output_dir if section is not None else ""
    # Set to standard error
    if log == "stderr" or (log is None and section is None):
        conf.log_settings.log = None
        if section == "pick":
            conf.pick.files.log = None
        elif section == "serialise":
            conf.serialise.files.log = None
        elif section == "prepare":
            conf.serialise.files.log = None
    # Set to the specified log output
    elif log is not None:
        if isinstance(log, (io.TextIOWrapper, io.BufferedWriter)):
            log.close()
            conf.log_settings.log = log.name
        elif isinstance(log, bytes):
            conf.log_settings.log = log.decode()
        elif isinstance(log, str):
            conf.log_settings.log = log
        else:
            raise TypeError("Invalid log type: {}".format(log))
    # infer from the section log, if present
    elif log is None and section is not None:
        other = getattr(conf, section).files.log
        if other is not None:
            conf.log_settings.log = other

    if conf.log_settings.log is not None:
        if not out_dir or (os.path.dirname(os.path.abspath(conf.log_settings.log)) == os.path.abspath(out_dir)):
            pass
        elif os.path.dirname(conf.log_settings.log) != "":
            conf.log_settings.log = os.path.abspath(conf.log_settings.log)
        elif os.path.dirname(os.path.abspath(conf.log_settings.log)) != os.path.abspath(out_dir):
            conf.log_settings.log = os.path.join(out_dir, conf.log_settings.log)

    if section == "pick":
        conf.pick.files.log = conf.log_settings.log
    elif section == "serialise":
        conf.serialise.files.log = conf.log_settings.log
    elif section == "prepare":
        conf.prepare.files.log = conf.log_settings.log

    try:
        conf.log_settings.log_level = log_level.upper() if log_level is not None else conf.log_settings.log_level
    except (TypeError, AttributeError):
        raise TypeError(f"Invalid log_level: {log_level}, type {type(log_level)}")
    # Check the log section
    errors = conf.log_settings.Schema().validate(dataclasses.asdict(conf.log_settings))
    if len(errors) > 0:
        raise InvalidConfiguration(
            f"Invalid logging selected for Mikado: log {log}, level {log_level}, section {section}\n"
            f"Errors: {errors}\nDict: {dataclasses.asdict(conf.pick.chimera_split)}")

    logger = create_logger_from_conf(conf, name=section if section else "mikado")
    return conf, logger



import re
import unittest

from .. import create_default_logger
from .._transcripts.scoring_configuration import ScoringFile, MinMaxScore, SizeFilter, RangeFilter
from ..loci import Transcript, Superlocus
import pkg_resources
import os
from ..configuration.configurator import load_and_validate_config
from ..serializers.external import ExternalSerializer
import pandas as pd


class ExternalTester(unittest.TestCase):

    def setUp(self):
        self.conf = load_and_validate_config(None)
        self.transcript = Transcript()
        self.transcript.chrom = "15"
        self.transcript.source = "protein_coding"
        self.transcript.start = 47631264
        self.transcript.end = 48051999

        exons = [(47631264, 47631416),
                 (47704590, 47704669),
                 (47762671, 47762742),
                 (47893062, 47893093),
                 (47895572, 47895655),
                 (48051942, 48051999)]

        self.transcript.strand = "+"
        self.transcript.add_exons(exons)
        self.transcript.id = "ENST00000560636"
        self.transcript.parent = "ENSG00000137872"
        self.transcript2 = self.transcript.copy()
        self.transcript2.id = "ENST00000560637"
        # self.assertIn("scoring", self.conf)
        
        # This is for the is_fragment test
        self.other_transcript = Transcript()
        self.other_transcript.start = 48053001
        self.other_transcript.end = 48053301

        exons = [(48053001, 48053301)]

        self.other_transcript.strand = "+"
        self.other_transcript.add_exons(exons)
        self.other_transcript.id = "ENST00000560646"
        self.other_transcript.parent = "ENSG00000137873"

    def test_copying(self):
        self.transcript.external_scores.update({"test": 0, "test1": 1})
        self.assertEqual(self.transcript.external_scores.test, 0)
        self.assertEqual(self.transcript.external_scores.test1, 1)
        transcript = self.transcript.deepcopy()
        self.assertEqual(transcript.external_scores.test, 0)
        self.assertEqual(transcript.external_scores.test1, 1)

    def test_no_handle(self):
        logger = create_default_logger("test_no_handle", level="DEBUG")
        with self.assertLogs(logger.name, level="WARNING") as cmo:
            ExternalSerializer(logger=logger, configuration=load_and_validate_config(None),
                               handle=None)
        self.assertTrue(any([record.msg == "No input file specified. Exiting." for record in cmo.records]))

    def test_no_fasta(self):
        conf = load_and_validate_config(None)
        conf.serialise.files.transcripts = "foo"
        logger = create_default_logger("test_no_fasta", level="DEBUG")
        with self.assertRaises(AssertionError), self.assertLogs(logger.name, level="CRITICAL") as cmo:
            ExternalSerializer(logger=logger, configuration=load_and_validate_config(None),
                               handle="trinity.bed12")
        self.assertTrue(any([record.msg == """I cannot find the mikado prepared FASTA file with the transcripts to analyse.
Please run mikado serialise in the folder with the correct files, and/or modify the configuration or the command line options."""
                             for record in cmo.records]))

    def test_invalid_tsv(self):
        conf = load_and_validate_config(None)
        conf.serialise.files.transcripts = pkg_resources.resource_filename("Mikado.tests",
                                                                           os.path.join("test_external",
                                                                                        "mikado_prepared.fasta"))
        self.assertTrue(os.path.exists(conf.serialise.files.transcripts),
                        conf.serialise.files.transcripts)
        logger = create_default_logger("test_no_fasta", level="DEBUG")
        h5 = pkg_resources.resource_filename("Mikado.tests", os.path.join("test_external", "abundance.h5"))
        with self.assertRaises((pd.errors.ParserError, UnicodeDecodeError)):
            ExternalSerializer(logger=logger, configuration=conf, handle=h5)
        wrong_tsv_one = pkg_resources.resource_filename("Mikado.tests", os.path.join("test_external", "wrong_1.tsv"))
        with self.assertRaises((ValueError, pd.errors.ParserError)), \
                self.assertLogs(logger.name, level="CRITICAL") as cmo:
            ExternalSerializer(logger=logger, configuration=conf, handle=wrong_tsv_one)
        self.assertTrue(any([re.search(r"Invalid input file", str(record.msg)) is not None for record in cmo.records]))
        wrong_tsv_two = pkg_resources.resource_filename("Mikado.tests", os.path.join("test_external", "wrong_2.tsv"))
        with self.assertRaises((ValueError, pd.errors.ParserError)), \
                self.assertLogs(logger.name, level="CRITICAL") as cmo:
            ExternalSerializer(logger=logger, configuration=conf, handle=wrong_tsv_two)
        self.assertTrue(any([re.search(r"Invalid input file", str(record.msg)) is not None for record in cmo.records]))

    def test_real(self):
        self.transcript.attributes["tpm"] = 10

        checked_conf = self.conf.copy()
        self.assertTrue(hasattr(self.conf.scoring.requirements, "parameters"))
        self.assertTrue(hasattr(checked_conf.scoring.requirements, "parameters"))
        checked_conf.scoring.scoring["attributes.tpm"] = MinMaxScore.Schema().load(
            {"rescaling": "max", "default": 0, "rtype": "float"})
        # self.assertIn('attributes.tpm', checked_conf.scoring.scoring)

        logger = create_default_logger(name="test_real", level="WARNING")
        sup = Superlocus(self.transcript, configuration=checked_conf, logger=logger)
        self.assertIn("attributes.tpm", sup.configuration.scoring.scoring)
        sup.get_metrics()
        self.assertIn("attributes.tpm", sup._metrics[self.transcript.id])
        self.assertEqual(sup._metrics[self.transcript.id]["attributes.tpm"], 10)

        sup2 = Superlocus(self.transcript2, configuration=checked_conf)
        sup2.get_metrics()
        self.assertIn("attributes.tpm", sup2._metrics[self.transcript2.id])
        self.assertEqual(sup2._metrics[self.transcript2.id]["attributes.tpm"], 0)

    def test_multiplier(self):

        self.transcript.attributes["tpm"] = 10

        checked_conf = self.conf.copy()
        checked_conf.scoring.scoring["attributes.tpm"] = MinMaxScore.Schema().load(
            {"rescaling": "max", "default": 0, "rtype": "float", 'multiplier': 4})

        self.assertIn('attributes.tpm', checked_conf.scoring.scoring)

        sup = Superlocus(self.transcript, configuration=checked_conf)
        tid = self.transcript.id
        self.assertIn(tid, sup.transcripts)
        sup.filter_and_calculate_scores(check_requirements=False)
        self.assertEqual(sup.scores[tid]['attributes.tpm'], 4)

    def test_default_attribute_score(self):
        self.transcript.attributes["foo"] = True

        checked_conf = self.conf.copy()
        checked_conf.scoring.scoring["attributes.foo"] = MinMaxScore.Schema().load(
            {"rescaling": "max", "rtype": "bool", "default": False})

        self.assertIn('attributes.foo', checked_conf.scoring.scoring)

        sup = Superlocus(self.transcript, configuration=checked_conf)
        sup.get_metrics()
        self.assertIn("attributes.foo", sup._metrics[self.transcript.id])
        self.assertEqual(sup._metrics[self.transcript.id]["attributes.foo"], True)

        sup2 = Superlocus(self.transcript2, configuration=checked_conf)
        sup2.get_metrics()
        self.assertIn("attributes.foo", sup2._metrics[self.transcript2.id])
        self.assertEqual(sup2._metrics[self.transcript2.id]["attributes.foo"], False)

    def test_error_attribute(self):
        self.transcript.attributes["tpm"] = "10a"
        checked_conf = self.conf.copy()
        checked_conf.scoring.scoring["attributes.tpm"] = MinMaxScore.Schema().load(
            {"rescaling": "max", "default": 0, "rtype": "float", 'multiplier': 4})
        self.assertIn('attributes.tpm', checked_conf.scoring.scoring)
        sup = Superlocus(self.transcript, configuration=checked_conf)
        with self.assertRaises(ValueError):
            sup.get_metrics()

    def test_attribute_use_raw(self):

        self.transcript.attributes["tpm"] = 10

        checked_conf = self.conf.copy()
        checked_conf.scoring.scoring["attributes.tpm"] = MinMaxScore.Schema().load(
            {"rescaling": "max", "default": 0, "rtype": "float", 'multiplier': 4,
             'use_raw': True})

        self.assertIn('attributes.tpm', checked_conf.scoring.scoring)

        sup = Superlocus(self.transcript, configuration=checked_conf)
        tid = self.transcript.id
        self.assertIn(tid, sup.transcripts)
        with self.assertRaises(ValueError):
            sup.get_metrics()

    def test_attribute_use_raw_percentage(self):

        self.transcript.attributes["tpm"] = 10

        checked_conf = self.conf.copy()
        checked_conf.scoring.scoring["attributes.tpm"] = MinMaxScore.Schema().load(
            {"rescaling": "max", "default": 0, "rtype": "float", 'multiplier': 4,
             'use_raw': True, 'percentage': True})

        self.assertIn('attributes.tpm', checked_conf.scoring.scoring)

        sup = Superlocus(self.transcript, configuration=checked_conf)
        tid = self.transcript.id
        self.assertIn(tid, sup.transcripts)
        sup.get_metrics()
        self.assertIn("attributes.tpm", sup._metrics[self.transcript.id])
        self.assertEqual(sup._metrics[self.transcript.id]["attributes.tpm"], 0.1)
        sup.define_loci()

    def test_attribute_use_as_alternative_splicing(self):
        from Mikado.loci import Locus
        checked_conf = self.conf.copy()
        self.transcript.attributes['cov'] = 10
        loci = Locus(self.transcript, configuration=checked_conf)
        loci.configuration.scoring.as_requirements.parameters['attributes.cov'] = SizeFilter(operator='ge', value=3,
                                                                                             metric=None,
                                                                                             name='attributes.cov')
        loci.configuration.scoring.as_requirements.expression = [
            loci.configuration.scoring.as_requirements.expression[0] + ' and attributes.cov']
        loci.configuration.scoring.as_requirements._check_my_requirements()
        self.assertTrue(loci._check_as_requirements(self.transcript))

    def test_attribute_use_as_alternative_splicing_fail_filter(self):
        from Mikado.loci import Locus
        checked_conf = self.conf.copy()
        self.transcript.attributes['cov'] = 10
        loci = Locus(self.transcript, configuration=checked_conf)
        loci.configuration.scoring.as_requirements.parameters['attributes.cov'] = SizeFilter(operator='ge', value=15,
                                                                                             metric=None,
                                                                                             name='attributes.cov')
        loci.configuration.scoring.as_requirements.expression = [
            loci.configuration.scoring.as_requirements.expression[0] + ' and attributes.cov']
        loci.configuration.scoring.as_requirements._check_my_requirements()
        self.assertFalse(loci._check_as_requirements(self.transcript))

    def test_attributes_use_as_cds_requirements(self):
        from Mikado.loci import Locus
        checked_conf = self.conf.copy()
        checked_conf.scoring.cds_requirements.expression = ["attributes.something"]
        checked_conf.scoring.cds_requirements.parameters = {"attributes.something": SizeFilter(value=1, operator="ge")}
        checked_conf.scoring.cds_requirements._check_my_requirements()
        self.transcript.attributes['something'] = 2
        loci = Locus(self.transcript, configuration=checked_conf)
        not_passing = loci._check_not_passing(section_name="cds_requirements")
        self.assertSetEqual(not_passing, set())

    def test_attributes_use_as_fragment_filter(self):
        from Mikado.loci import Locus
        checked_conf = self.conf.copy()
        checked_conf.pick.fragments.max_distance = checked_conf.pick.clustering.flank = 5000
        checked_conf.scoring.not_fragmentary.expression = ["attributes.something"]
        checked_conf.scoring.not_fragmentary.parameters = {"attributes.something": SizeFilter(value=1, operator="ge")}
        checked_conf.scoring.not_fragmentary._check_my_requirements()
        self.transcript.attributes['something'] = 2
        logger = create_default_logger("test_attributes_use_as_fragment_filter", level="DEBUG")
        loci = Locus(self.transcript, configuration=checked_conf, logger=logger)
        self.other_transcript.attributes["something"] = 0.5
        fragment = Locus(self.other_transcript, configuration=checked_conf, logger=logger)
        self.assertEqual(fragment.other_is_fragment(loci), (False, None))
        self.assertTrue(loci.other_is_fragment(fragment)[0])
        del self.other_transcript.attributes["something"]
        fragment = Locus(self.other_transcript, configuration=checked_conf, logger=logger)
        self.assertEqual(fragment.other_is_fragment(loci), (False, None))
        self.assertTrue(loci.other_is_fragment(fragment)[0])

    def test_attributes_range_use_as_fragment_filter(self):
        from Mikado.loci import Locus
        checked_conf = self.conf.copy()
        checked_conf.pick.fragments.max_distance = checked_conf.pick.clustering.flank = 5000
        checked_conf.scoring.not_fragmentary.expression = ["attributes.something"]
        checked_conf.scoring.not_fragmentary.parameters = {"attributes.something": RangeFilter(value=[10, 50],
                                                                                               operator="within")}
        checked_conf.scoring.not_fragmentary._check_my_requirements()
        self.transcript.attributes['something'] = 11
        logger = create_default_logger("test_attributes_range_as_fragment_filter", level="DEBUG")
        loci = Locus(self.transcript, configuration=checked_conf, logger=logger)
        self.other_transcript.attributes["something"] = 0.5
        fragment = Locus(self.other_transcript, configuration=checked_conf, logger=logger)
        self.assertEqual(fragment.other_is_fragment(loci), (False, None))
        self.assertTrue(loci.other_is_fragment(fragment)[0])
        del self.other_transcript.attributes["something"]
        fragment = Locus(self.other_transcript, configuration=checked_conf, logger=logger)
        self.assertEqual(fragment.other_is_fragment(loci), (False, None))
        self.assertTrue(loci.other_is_fragment(fragment)[0])
        # Now change the default.
        checked_conf.scoring.not_fragmentary.parameters = {"attributes.something": RangeFilter(value=[10, 50],
                                                                                               operator="within",
                                                                                               default=20)}
        checked_conf.scoring.not_fragmentary._check_my_requirements()
        not_fragment = Locus(self.other_transcript, configuration=checked_conf, logger=logger)
        self.assertNotIn("something", not_fragment.transcripts[self.other_transcript.id].attributes)
        # Not a fragment any more, the default of 15 is within the range.
        self.assertFalse(loci.other_is_fragment(fragment)[0])

    def test_print_metrics_with_attributes_from_scoring(self):
        from Mikado.loci import Locus, Sublocus
        checked_conf = self.conf.copy()
        checked_conf.scoring.scoring = {"attributes.something": MinMaxScore(default=0, rescaling="max", filter=None)}
        checked_conf.scoring.cds_requirements.expression = ["cdna_length"]
        checked_conf.scoring.cds_requirements.parameters = {"cdna_length": SizeFilter(operator="ge", value=1)}
        checked_conf.scoring.requirements.expression = ["cdna_length"]
        checked_conf.scoring.requirements.parameters = {"cdna_length": SizeFilter(operator="ge", value=1)}
        checked_conf.scoring.check(checked_conf.pick.orf_loading.minimal_orf_length)
        for lclass in [Locus, Sublocus]:
            logger = create_default_logger(f"test_print_metrics_with_attributes_{lclass.name}", level="DEBUG")
            locus = lclass(self.transcript, configuration=checked_conf, logger=logger)
            self.assertIn(self.transcript.id, locus.transcripts)
            rows = list(locus.print_metrics())
            self.assertEqual(len(rows), 1)
            row = rows[0]
            self.assertNotIn("something", locus.transcripts[self.transcript.id].attributes)
            self.assertIn("attributes.something", row.keys())
            self.assertEqual(row["attributes.something"], checked_conf.scoring.scoring["attributes.something"].default)
            self.transcript.attributes["something"] = 5
            locus = Locus(self.transcript, configuration=checked_conf)
            rows = list(locus.print_metrics())
            self.assertEqual(len(rows), 1)
            row = rows[0]
            self.assertIn("something", locus.transcripts[self.transcript.id].attributes)
            self.assertIn("attributes.something", row.keys())
            self.assertEqual(row["attributes.something"], 5)
            del self.transcript.attributes["something"]

    def test_print_metrics_with_attributes_from_requirements(self):
        from Mikado.loci import Locus, Sublocus
        checked_conf = self.conf.copy()
        checked_conf.scoring.scoring = {"attributes.something": MinMaxScore(default=0, rescaling="max", filter=None)}
        checked_conf.scoring.cds_requirements.expression = ["attributes.cds"]
        checked_conf.scoring.cds_requirements.parameters = {"attributes.cds": SizeFilter(operator="ge", value=1,
                                                                                         default=1)}
        checked_conf.scoring.requirements.expression = ["attributes.req"]
        checked_conf.scoring.requirements.parameters = {"attributes.req": SizeFilter(operator="ge", value=1,
                                                                                     default=1)}
        checked_conf.scoring.as_requirements.expression = ["attributes.as"]
        checked_conf.scoring.as_requirements.parameters = {"attributes.as": SizeFilter(operator="ge", value=1,
                                                                                       default=1)}
        checked_conf.scoring.not_fragmentary.expression = ["attributes.frag"]
        checked_conf.scoring.not_fragmentary.parameters = {"attributes.frag": SizeFilter(operator="ge", value=1,
                                                                                         default=1)}
        sections = {"something": checked_conf.scoring.scoring,
                    "req": checked_conf.scoring.requirements.parameters,
                    "as": checked_conf.scoring.as_requirements.parameters,
                    "cds": checked_conf.scoring.cds_requirements.parameters,
                    "frag": checked_conf.scoring.not_fragmentary.parameters}

        checked_conf.scoring.check(checked_conf.pick.orf_loading.minimal_orf_length)
        for lclass in [Locus, Sublocus]:
            logger = create_default_logger(f"test_print_metrics_with_attributes_{lclass.name}", level="DEBUG")
            locus = lclass(self.transcript, configuration=checked_conf, logger=logger)
            self.assertIn(self.transcript.id, locus.transcripts)
            rows = list(locus.print_metrics())
            self.assertEqual(len(rows), 1)
            row = rows[0]
            for key, section in sections.items():
                self.assertNotIn(key, locus.transcripts[self.transcript.id].attributes)
                self.assertIn(f"attributes.{key}", row.keys())
                self.assertEqual(row[f"attributes.{key}"],
                                 section[f"attributes.{key}"].default)
            for key in sections:
                self.transcript.attributes[key] = 5
            locus = Locus(self.transcript, configuration=checked_conf)
            rows = list(locus.print_metrics())
            self.assertEqual(len(rows), 1)
            row = rows[0]
            for key, section in sections.items():
                self.assertIn(key, locus.transcripts[self.transcript.id].attributes)
                self.assertIn(f"attributes.{key}", row.keys())
                self.assertEqual(row["attributes.something"], 5)
            # Reset the object
            for key in sections:
                del self.transcript.attributes[key]
